"""Set of functions used to test some assertions on datasets.
Useful when used in unit tests
"""

from collections.abc import Hashable, Iterable, Sequence

import numpy as np
import pandas as pd
from imageio.v3 import imread
from pandas.testing import assert_frame_equal
from tqdm.auto import tqdm

from ..dataset import Dataset
from ..utils import BBOX_COLUMN_NAMES

BOX_XMIN, BOX_YMIN, BOX_WIDTH, BOX_HEIGHT = BBOX_COLUMN_NAMES


def assert_column(
    input_df: pd.DataFrame,
    assertion: pd.Series | np.ndarray,
    message: str = "",
    n_first_occurrences: int | None = 1,
) -> None:
    """From a given input dataframe and a boolean series of the same length, construct
    an error message if the boolean has at least one False value, with the row in
    input dataframe corresponding to the row of the first occurrence of False value in
    the assertion series

    Args:
        input_df: Dataframe to show the row from, to better understand what went wrong
        assertion: Boolean Series of the same length as ``input_df``, expected to be
            full of True value
        message: Message to display when raising the error. Will be followed with
            information of faulty rows
        n_first_occurrences: Number of occurrences to show in case of a failure. Useful
            when showing duplicate values. If set to None, will show all occurrences.

    Raises:
        AssertionError: If there is at least one occurrence of False in ``assertion``
            Series, raise an assertion and print the corresponding row of first
            occurrence in ``input_df``
    """
    assert len(input_df) == len(assertion)
    assert n_first_occurrences is None or n_first_occurrences > 0
    assertion = assertion.astype(bool)
    if not assertion.all():
        failure = input_df[~assertion].iloc[:n_first_occurrences]
        if n_first_occurrences is None:
            raise AssertionError(f"Assertion failed. {message}. {failure}")
        elif n_first_occurrences == 1:
            raise AssertionError(
                f"Assertion failed. {message}. First occurrence at row"
                f" {failure.index[0]} : {failure.iloc[0]}"
            )
        else:
            raise AssertionError(
                f"Assertion failed. {message}. First occurrences at rows"
                f" {failure.index[:n_first_occurrences]} :\n{failure.iloc[:n_first_occurrences]}"
            )


def assert_columns_properly_normalized(
    input_df: pd.DataFrame, separator: str = "."
) -> None:
    """Checks that columns in input dataframes are well normalized, i.e. checks that
    if column 'A' exists, column 'A.B' does not exists.

    This is useful when loading json files to checks that a key cannot be both a
    sub dictionary and a value

    Args:
        input_df: Input DataFrame to test
        separator: Character used to separate name in flattened key. Defaults to ".".

    Raises:
        AssertionError: if there exist a column name where both the name and a variation
            of name + separator exists
    """
    for c in input_df.columns:
        prefix = f"{c}{separator}"
        for c2 in input_df.columns:
            if c2.startswith(prefix):
                raise AssertionError(
                    f"DataFrame is not properly normalized. Column '{c}' cannot be"
                    f" both a value and a subdictionary, but column '{c2}' exists"
                )


def assert_dataset_equal(
    dataset1: Dataset,
    dataset2: Dataset,
    ignore_index: bool = False,
    optional_columns: Iterable[str] = ("area", "confidence"),
    remove_na_columns: bool = False,
) -> None:
    """Compare two datasets and raise an assertion error if datasets are not equal.
    This function is mainly intended to be used in the context of unit tests.

    Rules:
        - Index order is not relevant. This is similar to ``check_like`` option
          in :func:`pandas.testing.assert_frame_equal`
        - Indexes for rows and columns still must be the same when reordered
        - Some columns in annotations are optional and are thus ignored if present in
          one but not the other dataset.
          If both are present, the columns' values are still compared.
        - Label maps must be the same. Again, order is ignored (as it normally is for
          dictionaries)
        - If ``ignore_index`` option is set to ``True``, index for rows are not checked,
          but we still check that the key in annotations' ``image_id`` points to the
          same rows in images dataframe

    Args:
        dataset1: First dataset to test
        dataset2: Second dataset to test, must be the same according to mentioned rules
            or the function will raise an error
        ignore_index: If set, will ignore both annotations and images dataframe index,
            but will still check that link between annotations and image row with
            ``image_id`` is the same. Defaults to False.
        optional_columns: Iterable of column names that will considered as optional,
            i.e. only check them if they are both present. Defaults to the column names
            "area" and "confidence".
        remove_na_columns: If set to True, will remove from dataframes columns where all
            values are equivalent to panda's ``<NA>``. This more lenient comparison is
            useful for columns where its absence and its values being all ``<NA>`` are
            treated the same, like the ``split`` column.

    Raises:
        AssertionError: raised when datasets are detected to be different
    """
    optional_columns = list(optional_columns)

    def assert_frame_equal_optional_columns(
        frame1: pd.DataFrame,
        frame2: pd.DataFrame,
        optional_columns: Sequence[str],
        dataframe_name: str,
    ) -> None:
        """Assert dataframe are equal, but first remove the optional columns if they are
        present in one dataframe and not in the other. Otherwise, if present in both,
        keep them for comparison
        """
        if remove_na_columns:
            frame1 = frame1.dropna(axis="columns", how="all")
            frame2 = frame2.dropna(axis="columns", how="all")
        for column_name in optional_columns:
            if column_name in frame1.columns and column_name not in frame2.columns:
                frame1 = frame1.drop(column_name, axis=1)
            if column_name in frame2.columns and column_name not in frame1.columns:
                frame2 = frame2.drop(column_name, axis=1)
        try:
            assert_frame_equal(frame1, frame2, check_like=True, check_dtype="equiv")
        except AssertionError as e:
            raise AssertionError(f"{dataframe_name} dataframes don't match") from e

    if ignore_index:
        dataset1 = dataset1.reset_index()
        dataset2 = dataset2.reset_index()

    assert_frame_equal_optional_columns(
        dataset1.images, dataset2.images, optional_columns, "Images"
    )
    assert_frame_equal_optional_columns(
        dataset1.annotations, dataset2.annotations, optional_columns, "Annotations"
    )
    assert (
        dataset1.label_map == dataset2.label_map
    ), f"label_maps don't match {dataset1.label_map} vs {dataset2.label_map}"

    assert dataset1.booleanized_columns == dataset2.booleanized_columns


def assert_frame_intersections_equal(df1: pd.DataFrame, df2: pd.DataFrame) -> None:
    """Construct inner dataframes from overlapping ids and columns and check they are
    equal

    These are the rows and columns present in both images dataframes
    The two dataframes must have the same values for the merge to be valid

    Args:
        df1: First dataframe to test
        df2: Second dataframe to test

    Raises:
        AssertionError: Raise error if both subdataframe constructed with intersections
            of indexes and columns are not the same.
    """
    df1_ids = set(df1.index)
    df2_ids = set(df2.index)
    mutual_ids = list(df1_ids & df2_ids)

    if not mutual_ids:
        return

    df1_columns = set(df1.columns)
    df2_columns = set(df2.columns)
    mutual_columns = list(df1_columns & df2_columns)

    if not mutual_columns:
        return

    inner_df1 = df1.loc[mutual_ids, mutual_columns]
    inner_df2 = df2.loc[mutual_ids, mutual_columns]
    try:
        assert_frame_equal(inner_df1, inner_df2)
    except AssertionError as e:
        raise AssertionError(
            "sub-Dataframes constructed from ids and columns in both DataFrames are not"
            " equal."
        ) from e


def assert_images_valid(
    dataset: Dataset,
    assert_is_symlink: bool = False,
    load_images: bool = True,
    check_exhaustive: bool = False,
) -> None:
    """Checks that the image paths in the dataset. Namely, checks that all path
    are indeed pointing to a file, and are valid file format that can be loaded with
    ``imageio``.

    Note:
        Todo: better error messages

    Args:
        dataset: Dataset to check
        assert_is_symlink: If set, will check that paths are symlinks rather than
            files. Defaults to False.
        load_images: If set to True, will not only check that images are valid files,
            but also that image can be loaded (i.e. are not corrupted files) and that
            their sizes match the ones included in ``dataset.images``
            dataframe. Note that this makes the function significantly slower.
            Defaults to True.
        check_exhaustive: If set to True, will check that all images in the images_root
            folder are in the image dataframe, and that the dataset is indeed exhaustive
    """
    get_invalid_images(
        dataset, assert_is_symlink, load_images, check_exhaustive, raise_if_error=True
    )


class InvalidImage(AssertionError):
    pass


class MissingImages(AssertionError):
    pass


def get_invalid_images(
    dataset: Dataset,
    check_symlink: bool = False,
    load_images: bool = True,
    check_exhaustive: bool = False,
    raise_if_error: bool = True,
) -> pd.DataFrame:
    """Checks dataset's images and return an indexed error report to retrieve them.

    Namely, checks that all path are indeed pointing to a file, and are valid file
    format that can be loaded with ``imageio``. If unsuccessful, add a row to the output
    dataframe with the same index as the faulty images, and info about the error in
    corresponding columns

    Args:
        dataset: Dataset to check
        check_symlink: If set, will check that paths are symlinks rather than
            files. Defaults to False.
        load_images: If set to True, will not only check that images are valid files,
            but also that image can be loaded (i.e. are not corrupted files) and that
            their sizes match the ones included in ``dataset.images``
            dataframe. Note that this makes the function significantly slower.
            Defaults to True.
        check_exhaustive: If set to True, will check that all images in the images_root
            folder are in the image dataframe, and that the dataset is indeed exhaustive
        raise_if_error: If set to True, will raise an InvalidImage error as soon as
            one image does not meet the requirements.

    Raises:
        InvalidImage: Raised if ``raise_if_error`` is selected and one image is not
            valid. Can be because the path is not right, the image loading failed,
            or the metadata is not compliant with actual image data.
        MissingImages: Raised if ``raise_if_error`` is selected and some images
            where found in the ``images_root`` folder but not in the dataset's
            ``images`` dataframe.

    Returns:
        Error report in the form of a Dataframe with "reason" and "additional_info"
        columns. Index values are the same as the corresponding images in the original
        dataset, so that you can retrieve the faulty images full data.
    """
    error_report = {}

    def error(
        message: str,
        additional_info: str,
        row_id: Hashable,
        image_data: "pd.Series[str]",
    ) -> None:
        error_report[row_id] = {"reason": message, "additional_info": additional_info}
        if raise_if_error:
            raise InvalidImage(
                f"{message}, {additional_info}\n row : {row_id}\n data:"
                f" {image_data.to_dict()}"
            )

    for row, img_data in tqdm(dataset.images.iterrows(), total=len(dataset)):
        if img_data["relative_path"].is_absolute():
            error("relative path is absolute", "", row, img_data)
            continue
        img_path = dataset.images_root / img_data["relative_path"]
        if check_symlink and not img_path.is_symlink():
            error("Not a symlink", "", row, img_data)
            continue
        valid_path = (
            img_path.is_symlink()
            and img_path.readlink().is_file()
            or img_path.is_file()
        )
        if not valid_path:
            error("Not a valid path", "", row, img_data)
            continue
        if load_images:
            try:
                img = imread(img_path)
            except OSError:
                error(
                    "corrupted file",
                    "Image cannot be loaded with imageio",
                    row,
                    img_data,
                )
                continue
            if not isinstance(img_data["width"], int) or img_data["width"] <= 0:
                error(
                    "Invalid image width",
                    f"got {img_data['width']} pixels",
                    row,
                    img_data,
                )
                continue
            if not isinstance(img_data["height"], int) or img_data["height"] <= 0:
                error(
                    "Invalid image height",
                    f"got {img_data['height']} pixels",
                    row,
                    img_data,
                )
                continue
            if len(img.shape) not in [2, 3, 4]:
                error(
                    "invalid image shape",
                    f"Shape is with {len(img.shape)} dimensions instead of 2"
                    " (grayscale), 3 (RGB/RGBA) or 4 (GIf anim)",
                    row,
                    img_data,
                )
            if len(img.shape) == 4:
                _, height, width, _ = img.shape
            elif len(img.shape) == 3:
                height, width, _ = img.shape
            else:
                height, width = img.shape
            if img_data["width"] != width:
                error(
                    "Image width in metadata is different from actual image width",
                    f"{width} (actual) vs {img_data['width']} (metadata)",
                    row,
                    img_data,
                )
            if img_data["height"] != height:
                error(
                    "Image height in metadata is different from actual image height",
                    f"{height} (actual) vs {img_data['height']} (metadata)",
                    row,
                    img_data,
                )
    if check_exhaustive:
        from ..dataset import from_folder

        highest_id = dataset.images.index.max()
        all_images = (
            from_folder(images_root=dataset.images_root)
            .reset_index(start_image_id=highest_id + 1)
            .images
        )
        missing_images = all_images.loc[
            ~all_images["relative_path"].isin(dataset.images["relative_path"]),
            "relative_path",
        ].apply(str)
        if len(missing_images) > 0 and raise_if_error:
            raise AssertionError(
                "Dataset is not exhaustive : the following images are present in"
                " images root but not in dataset image dataframe"
                f" :\n{', '.join(missing_images)}"
            )
        for row, relative_path in missing_images.items():
            error_report[row] = {
                "message": "missing image",
                "additional_info": relative_path,
            }

    return pd.DataFrame.from_dict(error_report, orient="index")


def assert_ids_well_formed(dataset: Dataset) -> None:
    """Assert ids follow the right convention.

    - DataFrames indexes must be named "id"
    - indexes must have no duplicates
    - images ``relative_path`` column must have no duplicates
    - annotation ``image_id`` values must all be in images index
    - annotation ``category_id`` values must be in dataset's label map

    Note:
        Todo: Better error messages

    Args:
        dataset: Dataset object to test.
    """
    assert dataset.images.index.name == "id", (
        "dataset's image index must be named 'id', got"
        f" {dataset.images.index.name} instead"
    )

    assert_column(
        dataset.images,
        ~dataset.images.index.duplicated(keep=False),
        "Dataset image index has duplicate values",
    )

    assert_column(
        dataset.images,
        ~dataset.images["relative_path"].duplicated(keep=False),
        "Dataset image relative path has duplicate values",
    )
    assert dataset.annotations.index.name == "id", (
        "dataset's annotation index must be named 'id', got"
        f" {dataset.annotations.index.name} instead"
    )
    assert_column(
        dataset.annotations,
        ~dataset.annotations.index.duplicated(keep=False),
        "Dataset annotations index has duplicate values",
    )
    assert_column(
        dataset.annotations,
        dataset.annotations["image_id"].isin(dataset.images.index),
        "All image_id values in annotations must in dataset images index",
    )
    assert_column(
        dataset.annotations,
        dataset.annotations["category_id"].isin(dataset.label_map.keys()),
        "All category ids must be in dataset label map",
    )


def assert_bounding_boxes_well_formed(
    dataset: Dataset, allow_keypoints: bool = False
) -> None:
    """Assert bounding boxes are well-formed in dataset's annotations.

    - Boxes x and y coordinates must be within their respective image size
    - Boxes width and height must be positive and so that xmax and ymin are within
      their respective image size
    - in the case of keypoints, Boxes with size 0 will be tolerated

    Args:
        dataset: Dataset to test
        allow_keypoints: If set to True, will not raise error if
            bounding box size (width or height) is 0. Defaults to False.
    """
    get_malformed_bounding_boxes(dataset, allow_keypoints, raise_if_error=True)


def get_malformed_bounding_boxes(
    dataset: Dataset, allow_keypoints: bool = False, raise_if_error: bool = False
) -> pd.DataFrame:
    """Get malformed bounding in dataset's annotations, as a boolean dataframe where
    index is id of bounding box in dataset's annotations dataframe, and columns are
    known reasons for bounding boxes to be invalid

    - Boxes x and y coordinates must be within their respective image size
    - Boxes width and height must be positive and so that xmax and ymin are within
      their respective image size
    - in the case of keypoints, Boxes with size 0 will be tolerated

    An invalid bounding box is then related to a row in the result dataframe where at
    least one of the value is True. Note that valid bounding boxes are NOT in the result
    dataframe. This means that if the dataset has no invalid bounding box, the result
    dataframe will be empty, and for each row in the result dataframe, there will be at
    least one ``True`` value.

    Args:
        dataset: Dataset to test
        allow_keypoints: If set to True, will not raise error if
            bounding box size (width or height) is 0. Defaults to False.
        raise_if_error: If set to True, will raise an error as soon as one bounding box
            is detected to be invalid. Defaults to False.

    Raises:
        AssertionError: When ``raise_if_error`` is set, raise an error as soon as one
            bounding box is invalid.

    Returns:
        Error report as a dataframe with boolean columns.

        - Each column is a reason why the bounding box can be faulty.
        - Each row is a faulty bounding box, with its corresponding index in dataset's
          annotation dataframe. Its value explain how the bounding box is invalid.
        - Only the faulty bounding boxes are kept in the error report, so all rows have
          at least one value set to True.
    """
    error_report = pd.DataFrame(index=dataset.annotations.index)

    def report_if_error(
        assertion: "pd.Series[bool]",
        error_name: str,
        message: str,
    ) -> None:
        error_report[error_name] = ~assertion
        if raise_if_error:
            assert_column(
                dataset.annotations, assertion, message, n_first_occurrences=1
            )

    report_if_error(
        dataset.annotations[BOX_XMIN] >= 0,
        "Negative X value",
        "Bounding boxes must have positive X values",
    )

    report_if_error(
        dataset.annotations[BOX_YMIN] >= 0,
        "Negative Y value",
        "Bounding boxes must have positive Y values",
    )
    report_if_error(
        dataset.annotations[BOX_WIDTH] >= 0,
        "Negative width",
        "Bounding boxes must have positive width",
    )
    report_if_error(
        dataset.annotations[BOX_HEIGHT] >= 0,
        "Negative height",
        "Bounding boxes must have positive height",
    )
    if not allow_keypoints:
        report_if_error(
            dataset.annotations[BOX_WIDTH] > 0,
            "0 width",
            "Bounding boxes must have strictly positive width",
        )
        report_if_error(
            dataset.annotations[BOX_HEIGHT] > 0,
            "0 height",
            "Bounding boxes must have strictly positive height",
        )
    x_max = dataset.annotations[BOX_XMIN] + dataset.annotations[BOX_WIDTH]
    y_max = dataset.annotations[BOX_YMIN] + dataset.annotations[BOX_HEIGHT]
    im_width = dataset.images.loc[dataset.annotations["image_id"], "width"]
    im_height = dataset.images.loc[dataset.annotations["image_id"], "height"]
    im_width.index = x_max.index
    im_height.index = x_max.index

    report_if_error(
        x_max <= im_width,
        "right side outside of image",
        "Bounding boxes must have X values below image width",
    )
    report_if_error(
        y_max <= im_height,
        "bottom side outside of image",
        "Bounding boxes must have Y values below image height",
    )
    return error_report[error_report.any(axis=1)]


def assert_label_map_well_formed(dataset: Dataset) -> None:
    """Assert label map has no category name duplicate

    Args:
        dataset: dataset to test.
    """
    label_map = pd.Series(dataset.label_map)
    assert (
        not label_map.duplicated().any()
    ), f"Label map dictionary has duplicate values : {dataset.label_map}"


def assert_required_columns_present(
    input_df: pd.DataFrame, required_columns: set[str], df_name: str
) -> None:
    """Simple function to check that required columns are present and raise a custom
    error if it's not the case

    Args:
        input_df: dataframe object to check.
        required_columns: set of column names to find in the columns of ``input_df``.
        df_name: name of the dataframe, used to add context to the error message.

    Raises:
        ValueError: Raised when not all required columns are present in the
            columns of ``input_df``.
    """
    missing_columns = required_columns - set(input_df.columns)
    if missing_columns:
        raise ValueError(
            f"DataFrame {df_name} must have all these columns"
            f" :\n{', '.join(required_columns)}\nbut is missing"
            f" {', '.join(missing_columns)}"
        )


def full_check_dataset_detection(
    dataset: Dataset,
    check_symlink: bool = False,
    allow_keypoints: bool = False,
    check_exhaustive: bool = False,
) -> None:
    """Perform a full check of the dataset. Images must be reachable for the test to
    perform.

    Args:
        dataset: dataset to test
        check_symlink: If set to True, will check that image relative paths are indeed
            relative links and not actual files. Defaults to False.
        allow_keypoints: If set to True, will not raise an error for bounding boxes with
            size 0 (width or height). Defaults to False.
        check_exhaustive: If set to True, will check that all images in the images_root
            folder are in the image dataframe, and that the dataset is indeed exhaustive
    """
    print("Checking Image and annotations Ids ...")
    assert_ids_well_formed(dataset)
    print("Checking Bounding boxes ..")
    assert_bounding_boxes_well_formed(dataset, allow_keypoints=allow_keypoints)
    print("Checking label map ...")
    assert_label_map_well_formed(dataset)
    print("Checking images are valid ...")
    get_invalid_images(
        dataset, check_symlink, True, check_exhaustive, raise_if_error=True
    )



import json
import shutil
from collections.abc import Iterable
from os.path import relpath
from pathlib import Path
from warnings import warn

import numpy as np
import pandas as pd
from imageio.v3 import imread, imwrite
from jsonschema_rs import JSONSchema
from tqdm.auto import tqdm

from lours.dataset import Dataset
from lours.utils import BBOX_COLUMN_NAMES
from lours.utils.testing import assert_columns_properly_normalized, assert_images_valid

from .common import construct_label_map
from .schema_util import (
    fill_with_dtypes_and_default_value,
    flatten_schema,
    get_enums,
    get_remapping_dict_from_names,
    get_remapping_dict_from_schema,
    load_json_schema,
    remap_dict,
)


def load_caipy_annot_folder(
    folder_path: Path, schema: dict | None = None
) -> tuple[pd.DataFrame | None, pd.DataFrame | None]:
    """Glob all json in folder path and construct image and annotation dataframe

    Args:
        folder_path: folder where we will search for json files
        schema: Optional JSON schema dict used to check the conformity of loaded
            JSON files.

    Returns:
        A pair of dataframes, representing image and annotations data,
        most likely used to construct the dataset object.
    """
    annotations = []
    images = []
    annotations_files = list(folder_path.glob("**/*.json"))
    if schema is not None:
        validator = JSONSchema(schema)
    else:
        validator = None
    for annot_file in tqdm(annotations_files):
        with open(annot_file) as f:
            frame_data = json.load(f)
        if validator is not None:
            validator.validate(frame_data)
        if "type" in frame_data.keys():
            assert (
                frame_data["type"] == "instances"
            ), "Only instance type supported for now"
        image_data = frame_data["image"]
        image_file_name = Path(image_data.pop("file_name"))
        assert annot_file.stem == image_file_name.stem, (
            "bad formatting, annotation file and image file_name should only differ by"
            f" the extension, got {annot_file.name} and"
            f" {image_file_name} instead"
        )
        image_path = annot_file.parent.relative_to(folder_path) / image_file_name
        image_data["relative_path"] = image_path
        image_data["type"] = image_path.suffix
        images.append(image_data)
        for annot in frame_data["annotations"]:
            annot["image_id"] = image_data["id"]
        annotations.extend(frame_data["annotations"])
    if images:
        # Sometimes, one key is either a sub dictionary or nan.
        # As a consequence, the column "key" will be full of nans while columns
        # "key.subkey" exist. remove columns full of NaN to avoid this situation.
        images = pd.json_normalize(images).set_index("id").dropna(axis=1, how="all")
        set_to_none_columns = images.select_dtypes("object").columns
        images[set_to_none_columns] = images[set_to_none_columns].replace(np.nan, None)
        assert_columns_properly_normalized(images)
    else:
        images = None
    if annotations:
        annotations = (
            pd.json_normalize(annotations).set_index("id").dropna(axis=1, how="all")
        )
        set_to_none_columns = annotations.select_dtypes("object").columns
        annotations[set_to_none_columns] = annotations[set_to_none_columns].replace(
            np.nan, None
        )
        # BBoxes are currently lists of 4 elements in a "bbox" column.
        # Convert each list element in the corresponding column for the dataset object
        bboxes = pd.DataFrame(list(annotations["bbox"]), index=annotations.index)
        annotations = annotations.drop("bbox", axis=1)
        for i, name in enumerate(BBOX_COLUMN_NAMES):
            annotations[name] = bboxes[i]
        assert_columns_properly_normalized(annotations)
    else:
        annotations = None

    return images, annotations


def load_caipy_split(
    images_folder: Path,
    annotations_folder: Path,
    dataset_name: str | None = None,
    split_name: str | None = None,
    schema: dict | None = None,
) -> Dataset:
    """Load a particular caipy split folder and convert it to a lours Dataset

    Args:
        images_folder: folder where images are stored
        annotations_folder: folder where annotations are stored as json files
        dataset_name: If specified, will be the dataset name, used when showing the
            dataset or exporting in other formats such as fiftyone. If not specified,
            the dataset name will be the name of the root folder.
        split_name: name of the split to give to ``split`` column of images DataFrame.
            Defaults to ``None``.
        schema: JSON schema dict used to check the conformity of loaded JSON files.
            If set to ``None``, will not check the conformity. Defaults to ``None``.

    Raises:
        ValueError: If image ids are not mutually exclusives

    Returns:
        Dataset containing only one split from caipy, expected to be merged with other
        caipy splits
    """
    images, annotations = load_caipy_annot_folder(annotations_folder, schema)
    if images is not None:
        if not images.index.is_unique:
            raise ValueError("two images have the same id")
        if split_name is not None:
            images["split"] = split_name

    label_map = (
        construct_label_map(annotations=annotations)
        if annotations is not None
        else None
    )
    return Dataset(
        images_root=images_folder,
        images=images,
        annotations=annotations,
        label_map=label_map,
        dataset_name=dataset_name,
    )


def from_caipy(
    dataset_path: Path | str,
    dataset_name: str | None = None,
    split: str | None = None,
    splits_to_read: str | Iterable[str] | None = None,
    use_schema: bool = False,
    json_schema: dict | str | Path | None = "default",
    booleanize: bool = True,
) -> Dataset:
    """Load a dataset stored in the cAIpy format

    See `specifications`__

    .. __: UPDATE-ME

    This will error if

    - two annotations have the same ``category_id`` but not the same ``category_str``
    - two annotations have a different ``category_id``  but the same ``category_str``
    - two images have the same ``file_name``, but not the same ``id``

    Args:
        dataset_path: folder root of dataset. Should contain the folders
            "Images" and "Annotations".
        dataset_name: If specified, will be the dataset name, used when showing the
            dataset or exporting in other formats such as fiftyone. If not specified,
            the dataset name will be the name of the root folder.
        split: if data is at the root of Images and Annotations folder, the split
            will be set to this option. Defaults to ``None``
        splits_to_read: if given, will only read the specified splits. Useful for a
            faster loading.
        use_schema: If set to True, and ``json_schema`` is not None, will use schema for
            validation and formatting (see option ``json_schema``)
        json_schema: schema dictionary or Path to a schema that json files will be
            tested against for compliance. If its not a dictionary, it can be either a
            url, or a local path. If set to None, or
            ``use_schema`` is set to False, will not perform any test.
            Defaults to default schema.
        booleanize: In the case some attributes are array of enum with unique
            elements, they will be booleanized
            (see :func:`~lours.dataset.Dataset.booleanize`).
            Note that this option is only used if `json_schema`` is not None and
            ``use_schema`` is set to True. Defaults to True.

    Raises:
        ValueError: Inconsistency between two annotations or images
            (see description above)


    Returns:
        Loaded dataset object


    See Also:
        - :func:`from_caipy_generic`
        - :ref:`Tutorial on schemas </notebooks/6_demo_schemas.ipynb>`
        - :ref:`Tutorial on booleanization </notebooks/7_demo_booleanize.ipynb>`
        - `cAIpy specifications <UPDATE-ME>`_
    """  # noqa: E501
    dataset_path = Path(dataset_path)
    annotations_folder = dataset_path / "Annotations"
    images_folder = dataset_path / "Images"
    dataset_name = dataset_name if dataset_name is not None else dataset_path.name
    return from_caipy_generic(
        images_folder,
        annotations_folder,
        dataset_name,
        split,
        splits_to_read,
        use_schema,
        json_schema,
        booleanize,
    )


def from_caipy_generic(
    images_folder: Path | str | None,
    annotations_folder: Path | str,
    dataset_name: str | None = None,
    split: str | None = None,
    splits_to_read: str | Iterable[str] | None = None,
    use_schema: bool = False,
    json_schema: dict | str | Path | None = "default",
    booleanize: bool = True,
) -> Dataset:
    """Load a dataset stored in the cAIpy format, but you can specify images and
    annotations folders rather than giving a folder with Images and Annotations
    sub-folders. This gives much more flexibility, especially when working predictions
    and annotations variations.

    See `specifications`__

    .. __: UPDATE-ME

    this will error if

    - two annotations have the same ``category_id`` but not the same ``category_str``
    - two annotations have a different ``category_id``  but the same ``category_str``
    - two images have the same ``file_name``, but not the same ``id``

    Args:
        images_folder: folder root of images.
        annotations_folder: folder root of annotations.
        dataset_name: If specified, will be the dataset name, used when showing the
            dataset or exporting in other formats such as fiftyone.
        split: if data is at the root of Images and Annotations folder, the split
            will be set to this option. Defaults to ``None``
        splits_to_read: if given, will only read the specified splits. Useful for a
            faster loading.
        use_schema: If set to True, and ``json_schema`` is not None, will use schema for
            validation and formatting (see option ``json_schema``)
        json_schema: schema dictionary or Path to a schema that json files will be
            tested against for compliance. If its not a dictionary, it can be either a
            url or a local path. If set to None, or
            ``use_schema`` is set to False, will not perform any test.
            Defaults to default schema.
        booleanize: In the case some attributes are array of enum with unique
            elements, they will be booleanized
            (see :func:`~lours.dataset.Dataset.booleanize`).
            Note that this option is only used if `json_schema`` is not None and
            ``use_schema`` is set to True. Defaults to True.

    Raises:
        ValueError: Inconsistency between two annotations or images
            (see description above)


    Returns:
        Loaded dataset object

    See Also:
        - :func:`from_caipy`
        - :ref:`Tutorial on schemas </notebooks/6_demo_schemas.ipynb>`
        - :ref:`Tutorial on booleanization </notebooks/7_demo_booleanize.ipynb>`
        - `cAIpy specifications <UPDATE-ME>`_
    """  # noqa: E501
    if use_schema and json_schema is not None:
        if isinstance(json_schema, dict):
            schema = json_schema
        else:
            schema = load_json_schema(json_schema)
    else:
        schema = None
    annotations_folder = Path(annotations_folder)
    if images_folder is not None:
        images_folder = Path(images_folder)
    else:
        images_folder = annotations_folder.parent / "Images"
        print(f"specifying a fictive path for images : {images_folder}")
    if isinstance(splits_to_read, str):
        splits_to_read = [splits_to_read]
    if splits_to_read is None:
        selected_splits = ["train", "valid", "eval"]
    else:
        selected_splits = splits_to_read

    dataset = Dataset(images_root=images_folder, dataset_name=dataset_name)
    for split_to_read in selected_splits:
        split_folder = annotations_folder / split_to_read
        if not split_folder.is_dir():
            continue
        split_dataset = load_caipy_split(
            images_folder=images_folder / split_to_read,
            annotations_folder=split_folder,
            dataset_name=dataset_name,
            split_name=split_to_read,
            schema=schema,
        )
        if len(split_dataset) == 0:
            continue
        dataset += split_dataset

    if len(dataset) == 0 and splits_to_read is None:
        dataset = load_caipy_split(
            images_folder=images_folder,
            annotations_folder=annotations_folder,
            dataset_name=dataset_name,
            split_name=split,
            schema=schema,
        )

    if schema is not None:
        image_schema = schema["properties"]["image"]
        annotation_schema = schema["properties"]["annotations"]["items"]
        if booleanize:
            image_enums = get_enums(image_schema)
            annotation_enums = get_enums(annotation_schema)
            dataset = dataset.booleanize(
                missing_ok=True, **(annotation_enums | image_enums)
            )
        dataset.images = fill_with_dtypes_and_default_value(
            image_schema, dataset.images
        )
        dataset.annotations = fill_with_dtypes_and_default_value(
            annotation_schema, dataset.annotations
        )
    return dataset


def split_to_caipy(
    dataset: Dataset,
    split_images_folder: Path | None,
    split_annotations_folder: Path,
    schema: dict | None = None,
    copy_images: bool = True,
    to_jpg: bool = True,
    overwrite_images: bool = True,
    overwrite_labels: bool = True,
    flatten_paths: bool = True,
) -> None:
    """Save a particular split to cAIpy. images and annotations folder must be given,
    as it can be the root of "Images" and "Annotations",
    or a subfolder based on split name, e.g. "Images/train"

    Note:
        Unless specified otherwise, relative paths of images a flattened during the
        export, which modifies the dataset if the images and annotations
        were stored in subfolders, but will put all images and annotations in their
        respective root folder.

    Note:
        If schema is not given, the nested dictionary will be deduced from column names
        with the separator "."

    Args:
        dataset: dataset object to save. Normally, should be a unique split
        split_images_folder: dataset where to save images, either as links or files.
            If None, will not save images. This is useful when you just want to save
            predictions or a variation of annotations.
        split_annotations_folder: dataset where to save caipyjson files.
        schema: JSON schema dict used to check the conformity of output JSON files.
            It will also be used to remove columns for fields no included
            in the schema. If set to ``None``, will not check the conformity.
            Defaults to ``None``.
        copy_images: If set to False, will create a symbolic link instead of copying.
            Much faster, but needs to keep original images in the same relative path.
            Defaults to False.
        to_jpg: if True, will convert images to jpg if needed. Defaults to True.
        overwrite_images: if set to False, will skip images that are already copied.
            Defaults to True.
        overwrite_labels: if set to False, will skip annotation that are already
            created. Defaults to True.
        flatten_paths: if set to True, will put all files in the root Annotations and
            Images folders by replacing folder separation ("/") with "_" in relative
            path. Defaults to True
    """
    if schema is not None:
        validator = JSONSchema(schema)
    else:
        validator = None
    # Get back to the list of 4 elements format
    converted_bboxes = pd.Series(
        dataset.annotations[BBOX_COLUMN_NAMES].to_numpy().tolist(),
        index=dataset.annotations.index,
        name="bbox",
    )
    # Remove useless columns and append the new one
    converted_annotations = dataset.annotations.drop(BBOX_COLUMN_NAMES, axis=1)
    converted_annotations = pd.concat([converted_annotations, converted_bboxes], axis=1)
    converted_annotations = converted_annotations.reset_index()
    # Get back the relative_path to a simple string
    n_images = len(dataset.images)
    if schema is None:
        image_remapping_dict = get_remapping_dict_from_names(
            frozenset(
                [*dataset.get_image_attributes(), "file_name", "id", "width", "height"]
            )
        )
        annotations_remapping_dict = get_remapping_dict_from_names(
            frozenset(
                [
                    *dataset.get_annotations_attributes(),
                    "id",
                    "bbox",
                    "category_id",
                    "category_str",
                ]
            )
        )
    else:
        image_schema = schema["properties"]["image"]
        annotations_schema = schema["properties"]["annotations"]["items"]
        image_remapping_dict = get_remapping_dict_from_schema(image_schema)
        saved_image_keys = flatten_schema(image_schema)
        lost_image_columns = set(dataset.get_image_attributes()) - set(saved_image_keys)
        if lost_image_columns:
            warn(
                "These column in self.images will be lost because they don't follow"
                f" the specified json schema: {', '.join(lost_image_columns)}",
                RuntimeWarning,
            )
        annotations_remapping_dict = get_remapping_dict_from_schema(annotations_schema)
        saved_annot_keys = flatten_schema(annotations_schema)
        lost_annot_columns = set(dataset.get_annotations_attributes()) - set(
            saved_annot_keys
        )
        if lost_annot_columns:
            warn(
                "These column in self.annotations will be lost because they don't"
                " follow the specified json schema:"
                f" {', '.join(lost_annot_columns)}",
                RuntimeWarning,
            )

    for image_id, image_data in tqdm(dataset.images.iterrows(), total=n_images):
        assert isinstance(image_id, int)
        instances = converted_annotations[converted_annotations["image_id"] == image_id]
        input_image_path = (dataset.images_root / image_data["relative_path"]).resolve()
        # Handle the case of images coming from a cAIpy, which already have the
        # structure {split}/filename in their relative path, which we don't want
        output_relative_path = image_data["relative_path"]
        if output_relative_path.parts[0] in ["train", "valid", "eval"]:
            output_relative_path = Path(*output_relative_path.parts[1:])
        if flatten_paths:
            output_relative_path = Path("_".join(output_relative_path.parts))
        output_filename = output_relative_path.name

        if split_images_folder is not None:
            output_image_path = (split_images_folder / output_relative_path).resolve()
            output_image_path.parent.mkdir(parents=True, exist_ok=True)
            if to_jpg:
                output_image_path = output_image_path.with_suffix(".jpg")
                output_filename = output_image_path.name
            # Copy image to destination. Do nothing if the image already exists,
            if overwrite_images or not output_image_path.is_file():
                if to_jpg and image_data["type"].lower() not in [".jpg", ".jpeg"]:
                    image = imread(input_image_path)
                    imwrite(output_image_path, image[..., :3])
                elif copy_images:
                    shutil.copy(input_image_path, output_image_path)
                else:
                    output_image_path.unlink(missing_ok=True)
                    output_image_path.symlink_to(
                        relpath(input_image_path, output_image_path.parent)
                    )

        output_json_path = split_annotations_folder / output_relative_path.with_suffix(
            ".json"
        )
        output_json_path.parent.mkdir(parents=True, exist_ok=True)
        if not output_json_path.is_file() or overwrite_labels:
            # Get only annotations associated with this image
            image_dict = {
                "file_name": output_filename,
                "id": int(image_id),
                "width": int(image_data["width"]),
                "height": int(image_data["height"]),
                **image_data[dataset.get_image_attributes()],
            }
            image_dict = remap_dict(image_dict, image_remapping_dict)
            annot_list = instances.to_dict("records")
            annot_list = list(
                map(lambda x: remap_dict(x, annotations_remapping_dict), annot_list)
            )
            annotations_dict = {
                "image": image_dict,
                "type": "instances",
                "annotations": annot_list,
            }
            if validator is not None:
                validator.validate(annotations_dict)

            with open(output_json_path, "w") as f:
                json.dump(annotations_dict, f, indent=2)


def dataset_to_caipy(
    dataset: Dataset,
    output_path: Path | str,
    use_schema: bool = False,
    json_schema: str | Path | None = "default",
    copy_images: bool = True,
    to_jpg: bool = True,
    overwrite_images: bool = True,
    overwrite_labels: bool = True,
    flatten_paths: bool = True,
) -> None:
    """Save dataset to cAIpy format
    Note that depending on the splits present in your dataset,
    the folder structure might change

    Note:
        Unless specified otherwise, relative paths of images a flattened during the
        export, which modifies the dataset if the images and annotations
        were stored in subfolders, but will put all images and annotations of a
        particular split in their respective root folder.

    Note:
        If schema is not given, the nested dictionary will be deduced from column names
        with the separator "."

    Args:
        dataset: dataset to save
        output_path: root folder where the dataset folder structure will be created.
        use_schema: If set to True, and ``json_schema`` is not None, will use schema for
            validation and formatting (see option ``json_schema``)
        json_schema: Path to a schema that output json dicts will be tested against for
            compliance. They will also be used to remove columns for fields no included
            in the schema. Can be either a url or a local path.
            If set to None, or ``use_schema`` is set to False,
            will not perform any test or reformatting. Defaults to default schema.
        copy_images: If set to False, will create a symbolic link instead of copying.
            Much faster, but needs to keep original images in the same relative path.
            Defaults to False.
        to_jpg: if True, will convert images to jpg if needed. Defaults to True.
        overwrite_images: if set to False, will skip images that are already copied.
            Defaults to True.
        overwrite_labels: if set to False, will skip annotation that are already
            created. Defaults to True.
        flatten_paths: if set to True, will put all files in the root Annotations and
            Images folders by replacing folder separation ("/") with "_" in relative
            path. Defaults to True

    """
    output_path = Path(output_path)
    images_folder = output_path / "Images"
    annotations_folder = output_path / "Annotations"
    return dataset_to_caipy_generic(
        dataset,
        images_folder,
        annotations_folder,
        use_schema,
        json_schema,
        copy_images,
        to_jpg,
        overwrite_images,
        overwrite_labels,
        flatten_paths,
    )


def dataset_to_caipy_generic(
    dataset: Dataset,
    output_images_folder: Path | str | None,
    output_annotations_folder: Path | str,
    use_schema: bool = False,
    json_schema: str | Path | None = "default",
    copy_images: bool = True,
    to_jpg: bool = True,
    overwrite_images: bool = True,
    overwrite_labels: bool = True,
    flatten_paths: bool = True,
) -> None:
    """Save dataset to cAIpy format
    Note that depending on the splits present in your dataset,
    the folder structure might change

    Notes:
        - Unless specified otherwise, relative paths of images a flattened during the
          export, which modifies the dataset if the images and annotations
          were stored in subfolders, but will put all images and annotations of a
          particular split in their respective root folder.
        - If schema is not given, the nested dictionary will be deduced from column names
          with the separator "."

    Args:
        dataset: dataset to save
        output_images_folder: root folder where the images will be saved. If None, will
            not save images. Useful when only saving predictions or a variations of
            annotations.
        output_annotations_folder: root folder where the json file will be saved.
        use_schema: If set to True, and ``json_schema`` is not None, will use schema for
            validation and formatting (see option ``json_schema``)
        json_schema: Path to a schema that output json dicts will be tested against for
            compliance. They will also be used to remove columns for fields no included
            in the schema. Can be either a url or a localt path.
            If set to None, or ``use_schema`` is set to False,
            will not perform any test. Defaults to the example schema.
        copy_images: If set to False, will create a symbolic link instead of copying.
            Much faster, but needs to keep original images in the same relative path.
            Defaults to False.
        to_jpg: if True, will convert images to jpg if needed. Defaults to True.
        overwrite_images: if set to False, will skip images that are already copied.
            Defaults to True.
        overwrite_labels: if set to False, will skip annotation that are already
            created. Defaults to True.
        flatten_paths: if set to True, will put all files in the root Annotations and
            Images folders by replacing folder separation ("/") with "_" in relative
            path. Defaults to True

    """
    if output_images_folder is not None:
        try:
            assert_images_valid(dataset, load_images=False)
        except AssertionError as e:
            raise ValueError(
                "Dataset images are missing, check that the images root folder is the"
                " right one"
            ) from e
    if use_schema and json_schema is not None:
        schema = load_json_schema(json_schema)
    else:
        schema = None
    output_dataset = dataset.debooleanize()
    if "split" in dataset.images.columns:
        splits = dataset.images["split"].unique().tolist()
    else:
        splits = []
    output_annotations_folder = Path(output_annotations_folder)
    if output_images_folder is not None:
        output_images_folder = Path(output_images_folder)
        output_images_folder.mkdir(exist_ok=True, parents=True)
    output_annotations_folder.mkdir(exist_ok=True, parents=True)
    # If no split, or only split evaluate to false (empty string or null value)
    # save the dataset without split subfolders
    if not splits or (len(splits) == 1 and pd.isna(splits[0])):
        print("Saving cAIpy dataset without split")
        split_to_caipy(
            output_dataset,
            output_images_folder,
            output_annotations_folder,
            schema,
            copy_images,
            to_jpg,
            overwrite_images,
            overwrite_labels,
            flatten_paths,
        )
    else:
        if any(pd.isna(split) for split in splits):
            raise AssertionError(
                "Dataset cannot have both data with known split and with unknown split"
            )
        for split_name in splits:
            print(f"Saving {split_name} split...")
            if output_images_folder is not None:
                split_images_folder = output_images_folder / f"{split_name}"
                split_images_folder.mkdir(exist_ok=True)
            else:
                split_images_folder = None
            split_annotations_folder = output_annotations_folder / f"{split_name}"
            split_annotations_folder.mkdir(exist_ok=True)
            split_to_caipy(
                output_dataset.get_split(split_name),
                split_images_folder,
                split_annotations_folder,
                schema,
                copy_images,
                to_jpg,
                overwrite_images,
                overwrite_labels,
                flatten_paths,
            )



from collections.abc import Iterable
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from lours.dataset import Dataset

from ...utils.grouper import get_group_names, group_list, group_relational_data
from .detection_evaluator_base import DetectionEvaluatorBase
from .util import (
    compute_average_precision,
    confusion_matrix,
    construct_matches_df,
    pr_curve,
)

if TYPE_CHECKING:
    pass


class DetectionEvaluator(DetectionEvaluatorBase):
    """Class specialization for detection tasks Note that the constructor is the
    same as the base Evaluator

    See Also:
        :ref:`related tutorial </notebooks/3_demo_evaluation_detection.ipynb>`
    """

    matches: dict[str, dict[str, pd.DataFrame]] = {
        "category_specific": {},
        "category_agnostic": {},
    }
    """Nested dictionary of DataFrames containing matched bounding boxes between
    groundtruth and corresponding prediction, depending on the way of computing
    matches (between all categories or between similar categories). Note that the
    the sub dictionaries will be empty until :func:`compute_matches` is called"""

    def __init__(
        self, groundtruth: Dataset, name: str | None = None, **predictions: Dataset
    ):
        """Constructor of the DetectionEvaluator class.
        The only difference with vanilla :class:`.Evaluator` is the matches that is set
        to an empty dictionary.
        """
        self.matches = {
            "category_specific": {},
            "category_agnostic": {},
        }
        super().__init__(groundtruth, name, **predictions)

    def compute_matches(
        self,
        predictions_names: str | Iterable[str] | None = None,
        min_iou: float = 0,
        category_agnostic: bool = False,
    ) -> dict[str, pd.DataFrame]:
        """Get matches between predictions and targets of the Evaluator.

        See Also:
            :ref:`Related tutorial </notebooks/3_demo_evaluation_detection.ipynb#Compute-the-matches>`

        Args:
            predictions_names: name or collection of prediction names to compute the
                matches on. If set to None, will compute the matches with the prediction
                DataFrames contained in the ``self.predictions_dictionary`` attribute.
                Defaults to None
            min_iou: IoU above which the detection is considered
                valid. Defaults to 0. Note that the lower bound of min_iou is not
                inclusive.
            category_agnostic: if set to False, matches are computed between categories,
                otherwise, matches are computed globally

        Returns:
            dict of DataFrame of matches, one entry per prediction specified in
            arguments. Will contain ``prediction_id`` and ``groundtruth_id`` columns.
            Index is unrelevant. Each prediction id and target id should appear once and
            only once. As such, at worse (no match at all), the dataframe will
            be :math:`N+M` rows with :math:`N` the number of predictions and :math:`M`
            the number of targets, and at best it will be :math:`max(M,N)`

        Example:
            >>> from lours.utils.doc_utils import dummy_dataset
            >>> groundtruth = dummy_dataset(
            ...     10,
            ...     1000,
            ...     label_map={0: "person", 1: "car"},
            ... )
            >>> predictions1 = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     add_confidence=True,
            ...     seed=0,
            ... )
            >>> predictions2 = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     add_confidence=True,
            ...     seed=1,
            ... )
            >>> evaluator = DetectionEvaluator(
            ...     groundtruth=groundtruth, A=predictions1, B=predictions2
            ... )
            >>> matches = evaluator.compute_matches()
            computing matches between groundtruth and A (category specific)
            computing matches between groundtruth and B (category specific)
            >>> matches["A"]
                  prediction_id       iou  groundtruth_id
            0             2311  0.370857             207
            1              515  0.586261             820
            2             7071  0.468022             585
            3             4444  0.089832              87
            4              235  0.431787             105
            ..             ...       ...             ...
            487           5016  0.000000            <NA>
            488           3608  0.000000            <NA>
            489            437  0.000000            <NA>
            490           8837  0.000000            <NA>
            491           2508  0.000000            <NA>
            <BLANKLINE>
            [10000 rows x 3 columns]

            You can select a particular set of prediction to only compute them

            >>> B_matches = evaluator.compute_matches(
            ...     predictions_names="B", category_agnostic=True
            ... )
            computing matches between groundtruth and B (category agnostic)
            >>> B_matches["B"]
                   prediction_id       iou  groundtruth_id
            0               7849  0.267152             832
            1               8819  0.089308             130
            2               6537  0.322729             785
            3               1616  0.406822             326
            4               8021  0.510778             929
            ...              ...       ...             ...
            1022            7377  0.000000            <NA>
            1023            8370  0.000000            <NA>
            1024            3534  0.000000            <NA>
            1025            7087  0.000000            <NA>
            1026            1410  0.000000            <NA>
            <BLANKLINE>
            [10000 rows x 3 columns]
        """
        if isinstance(predictions_names, str):
            predictions_names = [predictions_names]
        if predictions_names is None:
            predictions_names = list(self.predictions_dictionary)
        groundtruth = self.groundtruth.assign(groundtruth=True)
        matches = {}
        for prediction_name in predictions_names:
            print(
                f"computing matches between groundtruth and {prediction_name} (category"
                f" {'agnostic' if category_agnostic else 'specific'})"
            )
            prediction = self.predictions_dictionary[prediction_name].assign(
                groundtruth=False
            )
            groundtruth_prediction = pd.concat([groundtruth, prediction])

            groups = ["image_id"]
            tqdm.pandas()
            if category_agnostic:
                tqdm.pandas()
                matches = self.matches["category_agnostic"]
            else:
                matches = self.matches["category_specific"]
                groups.append("category_id")
            grouped = groundtruth_prediction.groupby(groups, group_keys=False)
            matches[prediction_name] = grouped.progress_apply(  # pyright: ignore
                partial(construct_matches_df, min_iou=min_iou), include_groups=False
            )
        return matches

    def compute_confusion_matrix(
        self,
        predictions_names: str | Iterable[str] | None = None,
        groups: group_list = (),
        min_iou: float = 0,
        min_confidence: float = 0,
    ) -> pd.DataFrame:
        """Compute confusion matrix to evaluate object detection.

        See Also:
            :ref:`related tutorial </notebooks/3_demo_evaluation_detection.ipynb#Computing-confusion-matrix>`

        Args:
            predictions_names: name or collection of prediction names to compute the
                matches on. If set to None, will compute the matches with the prediction
                DataFrames contained in the ``self.predictions_dictionary`` attribute.
                Default to None
            groups: Groups of image or annotation attributes to use to
                partition evaluation results to compute multiple confusion matrices.
                Must be a :obj:`.group_list` . Defaults to ()
            min_iou: IoU above which the detection is considered
                valid. Defaults to 0. Note that the lower bound of min_iou is not
                inclusive.
            min_confidence: confidence threshold above which the detection is considered
                valid. Defaults to 0. Note that the lower bound of min_confidence is not
                inclusive.

        Returns:
            A Dataframe with confusion data for each group name(s) (if any)
            of each predictions_names.

        Example:
            >>> from lours.utils.doc_utils import dummy_dataset
            >>> groundtruth = dummy_dataset(
            ...     10,
            ...     1000,
            ...     label_map={0: "person", 1: "car"},
            ... )
            >>> predictions1 = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     add_confidence=True,
            ...     seed=0,
            ... )
            >>> predictions2 = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     add_confidence=True,
            ...     seed=1,
            ... )
            >>> evaluator = DetectionEvaluator(
            ...     groundtruth=groundtruth, A=predictions1, B=predictions2
            ... )
            >>> evaluator.compute_confusion_matrix().reset_index().set_index(
            ...     ["model", "label"]
            ... )
            computing matches between groundtruth and A (category agnostic)
            computing matches between groundtruth and B (category agnostic)
            Processing confusion matrix for model=A
            Processing confusion matrix for model=B
                               car    person  None
            model label
            A     car     0.487179  0.512821   0.0
                  person  0.470588  0.529412   0.0
                  None    0.500889  0.499111   0.0
            B     car     0.495069  0.504931   0.0
                  person  0.470588  0.529412   0.0
                  None    0.505556  0.494444   0.0

            You can also use a minimum iou and select a subset of all prediction sets.

            >>> evaluator.compute_confusion_matrix(
            ...     min_iou=0.1, predictions_names="A"
            ... ).drop("model", axis=1)
            Processing confusion matrix for model=A
                         car    person      None
            label
            car     0.362919  0.374753  0.262327
            person  0.312373  0.367140  0.320487
            None    0.500377  0.499623  0.000000
        """
        if isinstance(predictions_names, str):
            predictions_names = [predictions_names]
        if predictions_names is None:
            predictions_names = list(self.predictions_dictionary)

        # compute category agnostic matches between prediction and groundtruth, and
        # construct a dataframe with groundtruth and prediction id, along with
        # category name for groundtruth (groundtruth_label) and prediction
        # (prediction_label). Groundtruth ids not matched to any prediction will have
        # a prediction label set to None (false negative) and prediction ids not matches
        # to any groundtruth will have a prediction label set to None (false positive).
        result_dict = {}
        group_names = get_group_names(groups)
        for name in predictions_names:
            if name not in self.matches["category_agnostic"]:
                self.compute_matches(name, category_agnostic=True)

            matches = self.matches["category_agnostic"][name]
            results = self.groundtruth.reset_index(names="groundtruth_id")
            results = pd.merge(
                results,
                matches.dropna(subset=["groundtruth_id"]),
                right_on="groundtruth_id",
                left_on="groundtruth_id",
            )
            results["confidence"] = 0.0
            results = results.rename(columns={"category_str": "groundtruth_label"})
            detected = results["prediction_id"].dropna()
            current_predictions = self.predictions_dictionary[name]
            results.loc[~results["prediction_id"].isna(), "confidence"] = (
                current_predictions.loc[detected, "confidence"].values
            )
            results.loc[~results["prediction_id"].isna(), "prediction_label"] = (
                current_predictions.loc[detected, "category_str"].values
            )

            # Then add the unmatched predictions
            false_positive = matches.loc[
                matches["groundtruth_id"].isna(), "prediction_id"
            ]
            results_fp = (
                current_predictions.loc[false_positive]
                .reset_index()
                .rename(
                    columns={"id": "prediction_id", "category_str": "prediction_label"}
                )
            )
            results_fp = pd.merge(
                results_fp,
                matches[matches["groundtruth_id"].isna()],
                right_on="prediction_id",
                left_on="prediction_id",
            )
            results = pd.concat([results, results_fp], ignore_index=True)

            # Keep matches with IOU above threshold or prediction and groundtruth which
            # have not been matched
            iou_above_threshold = (results["iou"] > min_iou) | (results["iou"] == 0)

            # Matches with IOU below threshold are duplicated to be considered
            # as False Negative and False Positive
            iou_below_threshold_gt = (results["iou"] <= min_iou) & results[
                ["groundtruth_label", "prediction_label"]
            ].notna().all(axis=1)
            iou_below_threshold_pred = iou_below_threshold_gt

            df_iou_above_threshold = results[iou_above_threshold]
            df_iou_below_threshold_gt = results[iou_below_threshold_gt].assign(
                prediction_label=pd.NA
            )
            df_iou_below_threshold_pred = results[iou_below_threshold_pred].assign(
                groundtruth_label=pd.NA
            )

            results = pd.concat(
                [
                    df_iou_above_threshold,
                    df_iou_below_threshold_gt,
                    df_iou_below_threshold_pred,
                ],
                ignore_index=True,
            )

            # Set predictions to None if confidence score lower or equal to
            # min_confidence argument
            results.loc[results["confidence"] <= min_confidence, "prediction_label"] = (
                pd.NA
            )
            # Remove rows where both prediction_label and groundtruth_label are None
            both_none_labels = (
                results[["prediction_label", "groundtruth_label"]].isna()
            ).all(axis=1)
            results = results[~both_none_labels]

            group_dict, category_groups, continuous_groups = group_relational_data(
                results, groups
            )
            pandas_groups = [group_dict[name] for name in group_names]
            result_dict[name] = (results, pandas_groups)

        confusion_dataframes = []
        for p_name, (current_results, pandas_groups) in result_dict.items():
            print(f"Processing confusion matrix for model={p_name}")
            if pandas_groups:
                grouped_data = current_results.groupby(
                    pandas_groups, observed=False
                ).apply(confusion_matrix)
                grouped_data["model"] = p_name
            else:
                grouped_data = confusion_matrix(current_results)
                grouped_data["model"] = p_name
            confusion_dataframes.append(grouped_data)
        return pd.concat(confusion_dataframes, ignore_index=False)

    def compute_precision_recall(
        self,
        predictions_names: str | Iterable[str] | None = None,
        groups: group_list = ("category_id",),
        ious: float | Iterable[float] = (0.0,),
        index_column: str | None = "recall",
        index_values: Iterable[float] | None = None,
        f_scores_betas: Iterable[float] = (1,),
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        r"""Compute Precision Recall curves, along with Average precision, with respect
        to recall, for different minimum IoU values.

        The dataset can be grouped, so that you get multiple pr curves in the end.

        It can be either groups of images (applied on ``self.images``) or groups of bbox
        (applied on ``self.groundtruth`` and ``self.predictions_dictionary``).

        In the case the data is not categorical, you must provide the number of
        desired bins of desired bin boundaries, and the cut method will be used to
        construct groups.

        See Also:
            - :ref:`Related tutorial </notebooks/3_demo_evaluation_detection.ipynb#Computing-AP-+-Yolov5-metrics>`
            - :class:`.ContinuousGroup`
            - :func:`pandas.cut` and :func:`pandas.qcut` for continuous groups.

        Note:
            For bbox groups, the value used will be the one of the target, except for
            false positive (no matching target) where the prediction data will be used.
            For example, the bbox size used for grouping will be the target one and not
            the prediction. So even if the prediction is out of bound, the detection
            will be considered valid as long as the IoU is high enough.
            However, when there is a false positive, the size of prediction will be used
            to decide in which group the precision needs to be decreased

        Args:
            predictions_names: names of predictions DataFrames, contained in
                ``self.predictions_dictionary`` to compute the PR curves on. If set to
                None, will compute PR curves for all predictions DataFrames.
            groups: Groups of image or annotation attributes to use to
                partition evaluation results to compute multiple PR curves. Must be a
                :obj:`.group_list` . Defaults to ``("category_id", )``.
            ious: minimum IoU values above which detection are considered valid.
                The higher, the harder it is for a detection to be valid.
                Defaults to 0.
            index_column: If set, will force the values of given column to be in the
                same bins. This will decrease data granularity, but make it possible to
                us this column as index. If not set, each category will have its own
                values, set exactly where recall and precision changes, making the curve
                more precise. Possible arguments are the only monotonous values (either
                increasing or decreasing), i.e. ``recall``, ``precision`` and
                ``confidence_threshold``. Defaults to ``recall`` to match pycocotools
                and fiftyone evaluation workflows.
            index_values: Iterable of bins, increasing float values from 0 to 1. used to
                reindex the dataframe. If set to None, will be 101 points evenly spaced
                from 0 to 1, to match pycocotools and fiftyone evaluation workflows.
                Defaults to None.
            f_scores_betas: beta values to compute the corresponding :math:`F_\beta`
                values in addition to precision and recall.

        Returns:
            PR curve dataset and corresponding average precision.
            The PR curve dataframe will have at least ``precision``,
            ``recall``, ``confidence_threshold``, and ``iou_threshold`` columns, plus
            the :math:`F_\beta` score columns, plus all the columns from the given
            groups. The AP dataframe will have at least AP and iou_threshold columns,
            plus all the columns from the given groups.

        Example:
            >>> from lours.utils.doc_utils import dummy_dataset
            >>> groundtruth = dummy_dataset(
            ...     10,
            ...     1000,
            ...     label_map={0: "person", 1: "car"},
            ...     n_attribute_columns_images={"attribute": 2},
            ... )
            >>> predictions1 = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     add_confidence=True,
            ...     seed=0,
            ... )
            >>> predictions2 = dummy_dataset(
            ...     10,
            ...     10000,
            ...     label_map=groundtruth.label_map,
            ...     images=groundtruth.images,
            ...     add_confidence=True,
            ...     seed=1,
            ... )
            >>> evaluator = DetectionEvaluator(
            ...     groundtruth=groundtruth, A=predictions1, B=predictions2
            ... )

            Get the Precision Recall curves and the Average Precision dataframe

            >>> pr, ap = evaluator.compute_precision_recall(ious=[0, 0.5])
            computing matches between groundtruth and A (category specific)
            computing matches between groundtruth and B (category specific)
            Processing PR curves for 2 IoU values and 2 prediction sets
            Processing PR curve for model=A and IOU=0
            Processing PR curve for model=B and IOU=0
            Processing PR curve for model=A and IOU=0.5
            Processing PR curve for model=B and IOU=0.5
            >>> ap
               category_id  iou_threshold model        AP category_str
            0            1            0.0     A  0.939509          car
            1            0            0.0     A  0.961933       person
            2            1            0.0     B  0.956845          car
            3            0            0.0     B  0.946684       person
            4            1            0.5     A  0.040764          car
            5            0            0.5     A  0.026722       person
            6            1            0.5     B  0.025094          car
            7            0            0.5     B  0.025750       person
            >>> pr
                 category_id  recall  precision  ...  iou_threshold  model  category_str
            0              1    0.00   1.000000  ...            0.0      A           car
            1              1    0.01   1.000000  ...            0.0      A           car
            2              1    0.02   1.000000  ...            0.0      A           car
            3              1    0.03   0.985714  ...            0.0      A           car
            4              1    0.04   0.985714  ...            0.0      A           car
            ..           ...     ...        ...  ...            ...    ...           ...
            803            0    0.96   0.000000  ...            0.5      B        person
            804            0    0.97   0.000000  ...            0.5      B        person
            805            0    0.98   0.000000  ...            0.5      B        person
            806            0    0.99   0.000000  ...            0.5      B        person
            807            0    1.00   0.000000  ...            0.5      B        person
            <BLANKLINE>
            [808 rows x 8 columns]

            For each class, iou and model, get the confidence threshold with the best f1
            and print the corresponding f1, recall and precision

            >>> best_f1 = pr.groupby(["model", "iou_threshold", "category_id"])[
            ...     "f1_score"
            ... ].idxmax()
            >>> pr.loc[best_f1, ["f1_score", "recall", "precision"]].set_index(
            ...     best_f1.index
            ... )
                                             f1_score  recall  precision
            model iou_threshold category_id
            A     0.0           0            0.904181    0.89   0.920502
                                1            0.884258    0.88   0.888889
                  0.5           0            0.131444    0.10   0.194030
                                1            0.158687    0.13   0.208202
            B     0.0           0            0.883191    0.87   0.903158
                                1            0.898718    0.90   0.898039
                  0.5           0            0.124986    0.10   0.168350
                                1            0.131136    0.10   0.203922

            Use a grouper to have PR values with respect to "attribute" image column,
            box height columns, thanks to a continuous group.

            >>> from lours.utils.grouper import ContinuousGroup
            >>> height_group = ContinuousGroup("box_height", bins=2, qcut=True)
            >>> pr, ap = evaluator.compute_precision_recall(
            ...     ious=0.1,
            ...     groups=["attribute", "category_id", height_group],
            ...     predictions_names="B",
            ... )
            Processing PR curves for 1 IoU value and 1 prediction set
            Processing PR curve for model=B and IOU=0.1
            >>> ap.set_index(["box_height", "attribute", "category_str"])["AP"]
            box_height          attribute  category_str
            (209.059, 938.398]  return     car             0.687178
            (0.0295, 209.059]   return     car             0.486098
                                to         person          0.394769
            (209.059, 938.398]  to         car             0.670351
                                return     person          0.727590
            (0.0295, 209.059]   to         car             0.517899
                                return     person          0.372749
            (209.059, 938.398]  to         person          0.586228
            Name: AP, dtype: float64


        """
        if isinstance(predictions_names, str):
            predictions_names = [predictions_names]
        if predictions_names is None:
            predictions_names = list(self.predictions_dictionary)
        for name in predictions_names:
            if name not in self.matches["category_specific"]:
                self.compute_matches(name, category_agnostic=False)

        if index_column is not None:
            if index_values is None:
                index_values = np.linspace(0, 1, 101)
            else:
                index_values = list(index_values)
            reindex_series = pd.Series(index_values, name=index_column)
        else:
            reindex_series = None

        assert self.matches

        result_dict = {}
        group_names = get_group_names(groups)
        for p_name in predictions_names:
            # Construct a dataframe of all targets, with corresponding matching
            # prediction (if any), and confidence from predictions (0 if false negative)
            matches = self.matches["category_specific"][p_name]
            current_predictions = self.predictions_dictionary[p_name]
            results = self.groundtruth.reset_index(names="groundtruth_id")
            results = pd.merge(
                results,
                matches.dropna(subset=["groundtruth_id"]),
                right_on="groundtruth_id",
                left_on="groundtruth_id",
            )
            results["confidence"] = 0.0
            detected = results["prediction_id"].dropna()
            results.loc[~results["prediction_id"].isna(), "confidence"] = (
                current_predictions.loc[detected, "confidence"].values
            )

            # Then add the unmatched predictions
            false_positive = matches.loc[
                matches["groundtruth_id"].isna(), "prediction_id"
            ]
            results_fp = (
                current_predictions.loc[false_positive]
                .reset_index()
                .rename(columns={"id": "prediction_id"})
            )
            results_fp = pd.merge(
                results_fp,
                matches[matches["groundtruth_id"].isna()],
                right_on="prediction_id",
                left_on="prediction_id",
            )
            results = pd.concat([results, results_fp], ignore_index=True)
            # groundtruth is True if groundtruth_id is NA, False otherwise
            results["groundtruth"] = ~results["groundtruth_id"].isna()

            results = results.sort_values("confidence", ascending=False)

            group_dict, *_ = group_relational_data(results, groups, self.images)
            pandas_groups = [group_dict[name] for name in group_names]
            result_dict[p_name] = (results, pandas_groups)

        precision_recall_curves = []
        if isinstance(ious, float | int):
            ious = [float(ious)]
        else:
            ious = [*ious]
        plural_ious = "s" if len(ious) > 1 else ""
        plural_predictions = "s" if len(result_dict) > 1 else ""
        print(
            f"Processing PR curves for {len(ious)} IoU value{plural_ious} and"
            f" {len(result_dict)} prediction set{plural_predictions}"
        )
        for iou in tqdm(ious):
            for p_name, (results, pandas_groups) in result_dict.items():
                print(f"Processing PR curve for model={p_name} and IOU={iou}")
                if pandas_groups:
                    precision_recall_curve = results.groupby(
                        pandas_groups, sort=False
                    ).progress_apply(  # pyright: ignore
                        partial(
                            pr_curve,
                            min_iou=iou,
                            reindex_series=reindex_series,
                            betas=f_scores_betas,
                        ),
                        include_groups=False,
                    )
                    # Groups are currently in the multiIndex. Reset it to make the
                    # dataframe easier to use: rename the index with the group names and
                    # make it dataframe columns.
                    # Also, the irrelevant id column at the same time
                    precision_recall_curve = precision_recall_curve.rename_axis(
                        [
                            *group_names,
                            "id",
                        ]
                    )
                    precision_recall_curve = precision_recall_curve.reset_index(
                        level=group_names
                    )
                else:
                    precision_recall_curve = pr_curve(
                        results, min_iou=iou, reindex_series=reindex_series
                    )
                precision_recall_curve["iou_threshold"] = iou
                precision_recall_curve["model"] = p_name
                precision_recall_curves.append(precision_recall_curve)

        precision_recall_curves = pd.concat(precision_recall_curves, ignore_index=True)
        average_precisions = precision_recall_curves.groupby(
            group_names + ["iou_threshold", "model"], sort=False
        ).apply(compute_average_precision, include_groups=False)
        average_precisions = average_precisions.rename("AP").reset_index()
        if "category_id" in group_names:
            precision_recall_curves["category_str"] = precision_recall_curves[
                "category_id"
            ].replace(self.label_map)
            average_precisions["category_str"] = average_precisions[
                "category_id"
            ].replace(self.label_map)
        return precision_recall_curves, average_precisions

    def _get_widgets(self):
        import ipywidgets as widgets
        from IPython.display import display

        components_widgets = super()._get_widgets()

        if self.matches:
            for category_name, matches_dict in self.matches.items():
                for prediction_name, matches in matches_dict.items():
                    widget_matches = widgets.Output()

                    # render in output widgets
                    with widget_matches:
                        display(matches)
                    components_widgets[
                        f"{category_name} {prediction_name} Matches (class)"
                    ] = widget_matches

        return components_widgets



from .dataset import Dataset
from .io.caipy import from_caipy, from_caipy_generic
from .io.coco import from_coco, from_coco_keypoints
from .io.crowd_human import from_crowd_human
from .io.darknet import (
    from_darknet,
    from_darknet_generic,
    from_darknet_json,
    from_darknet_yolov5,
)
from .io.images_folder import from_files, from_folder
from .io.mot import from_mot
from .io.parquet import from_parquet
from .io.pascalvoc import from_pascalVOC_detection, from_pascalVOC_generic

__all__ = [
    "Dataset",
    "from_folder",
    "from_files",
    "from_caipy",
    "from_caipy_generic",
    "from_coco",
    "from_coco_keypoints",
    "from_darknet",
    "from_darknet_yolov5",
    "from_darknet_generic",
    "from_darknet_json",
    "from_crowd_human",
    "from_mot",
    "from_parquet",
    "from_pascalVOC_detection",
    "from_pascalVOC_generic",
]



from collections.abc import Iterable, Sequence

import pandas as pd
from pandas._typing import Dtype


def reorder_columns(
    input_df: pd.DataFrame, first_columns: Sequence[str], separator: str = "."
) -> pd.DataFrame:
    """Reorder columns of input dataframe with a list of names and a separator character
    indicating what columns are nested

    The first columns will be the one specified in ``first_columns`` and the remaining
    columns will be reordered following alphabetical order, but nested columns (with a
    name that includes a ``separator`` character) are put at the end.

    Args:
        input_df: dataframe which columns are to be reordered
        first_columns: which columns to put first. Will preserve the order of the list,
            even if it's not sorted alphabetically.
        separator: Character used to separate column prefix and value. Used here to
            identify nested columns. Defaults to ".".

    Returns:
        Same dataframe as ``input_df``, except for the order of columns.
    """
    other_columns = set(input_df.columns) - set(first_columns)
    actual_first_columns = [name for name in first_columns if name in input_df.columns]

    def sort_columns(name: str) -> tuple[int, str]:
        return (name.count(separator), name)

    sorted_other_columns = sorted(other_columns, key=sort_columns)
    return input_df[[*actual_first_columns, *sorted_other_columns]]


def set_dataframe_dtypes(
    input_df: pd.DataFrame, dtypes_dict: dict[str, Dtype], nullable_types: Iterable[str]
) -> pd.DataFrame:
    """Set the right dtypes for columns of a DataFrame and a given name->dtype mapping.

    Notes:
        This script will ignore names which corresponding column is not present
        A list of column can be given so that they are deemed nullable. This is
            especially useful for string dtypes, because None is then converted to the
            'None' string.

    Args:
        input_df: input dataframe which columns are to be converted to desired dtype.
        dtypes_dict: dictionary with column names as key and the dtype as value. The
            dtype must be accepted by the :meth:`pandas.DataFrame.astype` method.
        nullable_types: list of columns which are nullable types. As such, for dtypes
            incompatible with null values (like string) the null objects converted to
            strings will be converted back to :obj:`pandas.NA`. Note that if you want
            to use pandas Nullable Integers, you need to explicitly give them in the
            ``dtypes_dict``.
            See https://pandas.pydata.org/docs/user_guide/integer_na.html

    Returns:
        same DataFrame as ``input_df``, except the dtypes of columns have been changed
        to the desired ones.
    """
    dtypes_to_assign = {
        name: dtype for name, dtype in dtypes_dict.items() if name in input_df.columns
    }
    actual_nullable_columns = [t for t in nullable_types if t in dtypes_to_assign]
    null_values = input_df[actual_nullable_columns].isna()
    output_df = input_df.astype(dtypes_to_assign)
    for column_name, column in null_values.items():
        output_df.loc[column, column_name] = pd.NA  # pyright: ignore
    return output_df



from typing import TYPE_CHECKING, Literal

from ..evaluator import Evaluator

if TYPE_CHECKING:
    import fiftyone as fo


class DetectionEvaluatorBase(Evaluator):
    """Base function for detection based evaluation. can be derived into basic
    detection or crowd detection
    """

    def to_fiftyone(
        self,
        name: str | None = None,
        record_fo_ids: bool = False,
        existing: Literal["error", "update", "erase"] = "error",
    ) -> "fo.Dataset":
        """Convert evaluator to fiftyone.

        Convert the detection evaluator into a
        :class:`fiftyone dataset <fiftyone.core.dataset.Dataset>`, that can then be
        inspected with Fiftyone's webapp. The resulting dataset will have the
        groundtruth sample field, along with all the prediction set's name and value in
        the ``self.predictions_dictionary`` attribute

        Args:
            name: Name of the fiftyone dataset to add the samples to. If the dataset
                does not exist, it will be created. If set to None, will use
                self.image_root folder name
            record_fo_ids: whether to record the fiftyone ids of samples and
                annotations. If set to True, will create ``fo_id`` column in self.images
                and ``fo_id`` and ``is_keypoint`` column in dataframes contained in
                self.predictions_dictionary and self.groundtruth to be able
                to reindex them in the created fiftyone dataset.
            existing: What to do in case there is already a fiftyone dataset with the
                same name.

                - "error": will raise an error.
                - "erase": will erase the existing dataset before uploading
                    this one
                - "update": will try to update the dataset by fusing together samples
                    with the same "relative_path"

                Defaults to "error".


        Returns:
            :class:`fiftyone.core.dataset.Dataset`: Fiftyone one dataset that can then
            be used to launch the webapp with
            ``fiftyone.launch_app(evaluator.to_fiftyone("dataset"))``

        See Also:
            - :ref:`Related tutorial </notebooks/5_demo_fiftyone.ipynb>`
            - :mod:`lours.utils.fiftyone_convert`
        """
        from ...utils.fiftyone_convert import create_fo_dataset

        if name is None:
            if self.name is None:
                name = self.images_root.name
            else:
                name = self.name

        fo_dataset, fo_image_ids, fo_annotations_ids = create_fo_dataset(
            name=name,
            images_root=self.images_root,
            images=self.images,
            annotations={
                "groundtruth": self.groundtruth,
                **self.predictions_dictionary,
            },
            label_map=self.label_map,
            image_tag_columns=self.get_image_attributes(),
            annotations_attributes_columns={
                "groundtruth": self.get_annotations_attributes(),
                **{
                    name: self.get_annotations_attributes(name)
                    for name in self.predictions_dictionary
                },
            },
            existing=existing,
        )

        if record_fo_ids:
            self.images["fo_id"] = fo_image_ids
            for name, predictions_frame in self.predictions_dictionary.items():
                # a simple concat with axis=1 could work, but this snippet is clearer
                # in terms of what column are expected and acts as a check
                predictions_frame["fo_id"] = fo_annotations_ids[name]["fo_id"]
                predictions_frame["is_keypoint"] = fo_annotations_ids[name][
                    "is_keypoint"
                ]
            self.groundtruth["fo_id"] = fo_annotations_ids["groundtruth"]["fo_id"]
            self.groundtruth["is_keypoint"] = fo_annotations_ids["groundtruth"][
                "is_keypoint"
            ]

        return fo_dataset



class IncompatibleLabelMapsError(ValueError):
    """Custom Exception to differentiate ValueError because of incompatible label maps
    from other ValueError exceptions
    """

    pass


def merge_label_maps(
    left: dict[int, str], right: dict[int, str], method: str = "inner"
) -> dict[int, str]:
    """Merge two label maps together, either by trying to include one in the other
    with the inner method, or by constructing the union of the two with the outer method

    Args:
        left: First label map to merge
        right: Second label map to merge
        method: use inner label map or outer

    Raises:
        IncompatibleLabelMapsError: Raised when label maps are incompatible. Namely,
            there exists a ``category_id`` in both label maps, but they have different
            names. In the case of inner method, a IncompatibleLabelMapsError can also be
            raised if the smallest label map is not included in the other.

    Returns:
        Merged label map

    Example:
        >>> label_map1 = {0: "car", 1: "person", 2: "truck"}
        >>> label_map2 = {0: "car", 1: "person"}
        >>> merge_label_maps(label_map1, label_map2)
        {0: 'car', 1: 'person', 2: 'truck'}

        >>> label_map2[3] = "head"
        >>> merge_label_maps(label_map1, label_map2)
        Traceback (most recent call last):
            ...
        lours.utils.label_map_merger.IncompatibleLabelMapsError: Label maps are incompatible

        >>> merge_label_maps(label_map1, label_map2, method="outer")
        {0: 'car', 1: 'person', 2: 'truck', 3: 'head'}

        >>> label_map2[0] = "vehicle"
        >>> merge_label_maps(label_map1, label_map2, method="outer")
        Traceback (most recent call last):
            ...
        lours.utils.label_map_merger.IncompatibleLabelMapsError: Label maps are incompatible
    """
    # Check that the two label maps are compatible.
    # If not you will have to remap one of them
    if len(left) < len(right):
        left, right = right, left
    if method == "inner":
        # If this dataset's object map is the biggest, the other one must be
        # a subset of it (keys and values)
        if not set(right).issubset(set(left)):
            raise IncompatibleLabelMapsError("Label maps are incompatible")
        if right != {k: left[k] for k in right}:
            raise IncompatibleLabelMapsError("Label maps are incompatible")
        return left
    if method == "outer":
        intersection = set(left) & set(right)
        # The other way around when the other dataset's label map is the biggest
        if {k: left[k] for k in intersection} != {k: right[k] for k in intersection}:
            raise IncompatibleLabelMapsError("Label maps are incompatible")
        return left | right
    else:
        raise ValueError("method can only be 'inner' or 'outer'")



from typing import Any

import numpy as np
import pandas as pd
from ot import dist, emd2, sinkhorn2
from scipy.stats import entropy


def earth_mover_distance(
    left: pd.Series,
    right: pd.Series,
    continuous_weights: pd.Series,
    sinkhorn_lambda: float = 0,
) -> float:
    """Compute earth mover distance between two columns of a dataframe.

    Note:
        In the case of ``sinkhorn_lambda`` > 0 this uses the sinkhorn algorithm for a
        faster approximate value.

        See :func:`ot.sinkhorn2`

    Args:
        left: input Series that represents histograms (not necessarily normalized), and
            the index represent the histogram bins
        right: input Series that represents histograms (not necessarily normalized), and
            the index represent the histogram bins. Note that ``left`` and ``right``
            don't necessarily share the same bins.
        continuous_weights: Series of index level names to consider in the
            ``left_right_df`` dataframe for the sinkhorn algorithm.
        sinkhorn_lambda: regularization weight for sinkhorn algorithm. If 0, will use
            literal earth mover distance without regularization (slower but more
            accurate). Defaults to 0.

    Returns:
        distance between the two histograms
    """
    if len(continuous_weights) == 0:
        return 0
    left_sum, right_sum = left.sum(), right.sum()
    if left_sum == 0 or right_sum == 0:
        return float("inf")
    left = left[left > 0]
    right = right[right > 0]
    left_bins = (
        left.index.to_frame()[continuous_weights.index].astype(float)
        * continuous_weights
    ).to_numpy()
    right_bins = (
        right.index.to_frame()[continuous_weights.index].astype(float)
        * continuous_weights
    ).to_numpy()
    distance_matrix = dist(left_bins, right_bins, metric="cityblock")
    distance_matrix = distance_matrix / distance_matrix.max()

    normalized_left = (left / left.sum()).to_numpy()
    normalized_right = (right / right.sum()).to_numpy()
    if sinkhorn_lambda == 0:
        return emd2(
            normalized_left, normalized_right, distance_matrix
        )  # pyright: ignore
    else:
        return sinkhorn2(
            normalized_left, normalized_right, distance_matrix, reg=sinkhorn_lambda
        )  # pyright: ignore


def check_groups(
    histogram: pd.DataFrame | pd.Series,
    category_groups: pd.Series,
    continuous_groups: pd.Series,
) -> None:
    """Check that histogram and groups are well-formed.

    Namely:
     - There should be no overlap between the two groups
     - histogram must have as many index dimensions as the total number of groups
     - histogram multi-index names must be unique
     - there should be a bijection between histogram index names and given category
       and continuous groups


    Args:
        histogram: Series or DataFrame with one or two columns, and a multi index whose
            names must match the next two groups
        category_groups: Series whose index are names of category groups, which should
            be contained in the histogram index
        continuous_groups: Series whose index are names of continuous groups, which
            should be contained in the histogram index

    Raises:
        AssertionError: raises an error when histogram and groups don't respect
            aforementioned criteria
    """
    if len(category_groups) == 0 and len(continuous_groups) == 0:
        raise AssertionError("no group to use pandas' groupby on")
    elif len(category_groups) == 0:
        total_groups = continuous_groups.index
    elif len(continuous_groups) == 0:
        total_groups = category_groups.index
    else:
        total_groups = pd.concat([category_groups, continuous_groups]).index
    assert (
        total_groups.is_unique
    ), "category and continuous groups must have unique and non overlapping values"

    histogram_names = set(histogram.index.names)
    assert len(histogram_names) == len(
        histogram.index.names
    ), "histogram multi index must have index with exclusive names"
    assert histogram_names == set(total_groups), (
        "category and continuous group must be a "
        "perfect partition of the histogram index"
    )
    return


def hist_distance(
    left: pd.Series,
    right: pd.Series,
    category_weights: pd.Series,
    continuous_weights: pd.Series,
    sinkhorn_lambda: float = 0,
) -> float:
    r"""Compute the distance between two distributions described in pandas Series
    representing histograms. Both index must match and may have categorical data or
    continuous data. Distance between categorical data is made with
    Kullback–Leibler divergence and distance between continuous data us made with
    Earth mover distance.

    the distance formula is then

    .. math::
        :label: hist_cost

        D = \sum_{0 \le i < p} \alpha_i KL\left(
            P_{cat, C_i}, Q_{cat, C_i}
            \right) + || \beta || \sum_{i \in \Omega_{cat}} \left(
                P_{cat}(i) \times EMD(P^\beta(i), Q^\beta(i))
            \right)

    where
     - :math:`p \in \mathbb{N}` and :math:`q \in \mathbb{N}` are respectively
       the number of categorical dimensions and continuous dimensions
     - :math:`\Omega_{cat} \subset \mathbb{N}^p` is the set of all possible
       categories, subdivided into :math:`p` dimensions

       .. math::

           \Omega_{cat} &=
               \{ c_{0,0}, c_{1,0} \cdots,  c_{n_0, 0} \}
               \times \cdots
               \times \{ c_{0, p}, \cdots, c_{n_p, p} \} \\
            \Omega_{cat} &=
               C_0 \times \cdots \times C_p

     - :math:`P` is the probability function of the histogram

       .. math::

           P : \begin{array}{lll}
              \Omega_{cat} \times \mathbb{R}^q & \rightarrow & [ 0, 1 ] \\
              (x,y) = (x_0, \cdots, x_p, y_0 \cdots y_p) & \mapsto & P(x,y)
           \end{array}

     - :math:`P_{cat}` is the agglomeration of :math:`P` over continuous dimensions.

       .. math::

           P_{cat} :
               \begin{array}{lll}
                   \Omega_{cat} & \rightarrow & [0, 1] \\
                   x & \mapsto & \iint_{y \in \mathbb{R}^q} P(x, y) dy
                \end{array}

     - :math:`P_{cat, C_i}` is the agglomeration of :math:`P` over continuous
       dimensions and category dimensions except :math:`C_i`

       .. math::

           P_{cat, C_i} &: C_i \rightarrow [0, 1]

           P(x) &= \sum_{
               x' \in C_0 \times \cdots \times C_{i-1} \times C_{i+1}
               \times \cdots \times C_p
           } \iint_{y \in \mathbb{R}^q}
               P(x'_0, \cdots x'_{i-1}, x, x'_{i+1} \cdots x'_p, y) dy

     - :math:`P(x)` is the probability distribution over continuous dimensions for a
       particular category :math:`x \ in \Omega_{cat}`.

       .. math::

            P(x) :
               \begin{array}{lll}
                   \mathbb{R}^q & \rightarrow & [0, 1] \\
                   y & \mapsto & \  P(x, y)
                \end{array}
     - :math:`P^\beta(x)` is the weighted probability distribution over continuous
       dimensions for a particular class :math:`x` and a weight vector :math:`\beta`

       .. math::

           P^\beta(x) &: \mathbb{R}^q \rightarrow [0, 1]

           P^\beta(x,y) &= P \left(x, \frac{\beta}{|| \beta ||} \odot y\right)

     - :math:`\alpha \in \mathbb{R}^p` and
       :math:`\beta \in \mathbb{R}^q` are weight vectors associated to importance
       of each dimensions of :math:`\Omega_{cat} \times \mathbb{R}^q`
     - :math:`\odot` is the Hadamard product

       .. math::

           \beta \odot y = (\beta_j y_j)_{0 \le j < p}
     - :math:`KL` is the Kullback–Leibler divergence
     - :math:`EMD` is the Earth Mover distance

    Note:
        This formula is not symmetric, it is more suited to compare a reference
        distribution (the left one) to a candidate distribution (the right one).


    Args:
        left: pandas Series representing left distribution of probability
            (i.e. the reference)
        right: pandas Series representing left distribution of probability
            (i.e. the candidate)
        category_weights: weights Series vector associated with :math:`\alpha` which is
            applied to the KL divergence (see formula :eq:`hist_cost`). Its index must
            be the names of category groups, that represent ``left`` and ``right``
            indexes dimensions on which to apply KL divergence.
        continuous_weights: weight Series vector associated with :math:`\beta` which is
            applied to the Earth mover's distance (see formula :eq:`hist_cost`).
            Its index must be the names of category groups, that represent
            ``left`` and ``right`` indexes dimensions on which to apply EMD.
        sinkhorn_lambda: regularization term applied to EMV
            (see :func:`earth_mover_distance`). Defaults to 0


    Returns:
        distance between the two multimodal distributions.
    """
    left_right = pd.concat([left, right], axis="columns", keys=["left", "right"])
    check_groups(left_right, category_weights, continuous_weights)
    if 0 in left_right.sum().values:
        return float("inf")

    kl_div = 0
    if len(category_weights) > 0:
        by_cat = left_right.groupby(list(category_weights.index), observed=True)
        by_cat_count = by_cat.sum()

        for axis, weight in category_weights.items():
            assert isinstance(axis, str)
            by_cat_count = left_right.groupby(axis, observed=True).sum()
            axis_entropy = entropy(
                by_cat_count["left"].values, by_cat_count["right"].values
            )
            kl_div += weight * axis_entropy

        emd_by_cat = by_cat.apply(
            lambda left_right_df: earth_mover_distance(
                left=left_right_df["left"],
                right=left_right_df["right"],
                continuous_weights=continuous_weights,
                sinkhorn_lambda=sinkhorn_lambda,
            )
        )
        emd = (by_cat_count["left"] * emd_by_cat).sum() / by_cat_count["left"].sum()
    else:
        emd = earth_mover_distance(
            left=left_right["left"],
            right=left_right["right"],
            continuous_weights=continuous_weights,
            sinkhorn_lambda=sinkhorn_lambda,
        )
    return (
        np.linalg.norm(category_weights.to_numpy()) * kl_div  # pyright: ignore
        + np.linalg.norm(continuous_weights.to_numpy()) * emd
    )


def df_to_hist(
    data: pd.DataFrame,
    groupby: Any,
    full_index: pd.Index | pd.MultiIndex | None = None,
) -> pd.Series:
    """Convert dataframe to histograms by using pandas'
    `GroupBy <https://pandas.pydata.org/docs/reference/groupby.html>`__ feature

    Args:
        data: DataFrame from which the histogram will be computed. Must have the columns
            specified in groups option.
        groupby: Same ``by`` option for :meth:`pandas.DataFrame.groupby`, will be passed
            directly to ``data.groupby`` method. Can be a mapping, a function, a label,
            or a list of labels.
        full_index: Optional index to reindex the resulting histogram. Useful when some
            value have an occurrence count of 0 and thus don't appear in the induced
            index. Defaults to None.

    Returns:
        pandas Series with multiindex corresponding to the count of occurrences for each
        specified group.
    """
    hist = data.groupby(groupby, observed=False).size()
    hist.name = "histogram"
    if full_index is None:
        if isinstance(hist.index, pd.MultiIndex):
            full_index = pd.MultiIndex.from_product(hist.index.levels)
        else:
            full_index = hist.index
    return hist.reindex(full_index).fillna(0)


def dataset_share_distance(left_share: pd.Series, right_share: pd.Series) -> float:
    r"""Compute the distance between two dataset share histograms (where bins are
    splits) by using Intersection over Union (IoU). We use this distance instead of KL
    because we don't want an infinite distance when one of the split is empty.

    .. math::
        D = \frac{\sum_{i=0}^{n_{splits}} min(left(i), right(i))}
                  {\sum_{i=0}^{n_{splits}} max(left(i), right(i))}

    Args:
        left_share: Series representing target histogram of split sizes.
            It has to be normalized.
        right_share: candidate histogram of split sizes

    Returns:
        distance computed
    """
    left_share = left_share / left_share.sum()
    right_share = right_share / right_share.sum()
    intersection = np.minimum(left_share, right_share)
    union = np.maximum(left_share, right_share)
    if union.sum() == 0:
        print(left_share, right_share)
        raise ValueError()
    return 1 - intersection.sum() / union.sum()



from pathlib import Path

from lours.dataset import Dataset
from lours.utils import parquet_saver


def from_parquet(input_path: Path | str) -> Dataset:
    """Load a Dataset object from a folder with parquet files for its dataframes.

    Other attributes will be loaded from dataset.yaml file in the same folder.

    Args:
        input_path: Folder to read the yaml file and parquet files from.

    Raises:
        ValueError: Raised when the object name contained in
            ``input_dict['__name__']`` is not 'Dataset'.

    Returns:
        Loaded dataset
    """
    data = parquet_saver.dict_from_parquet(Path(input_path))
    if data["__name__"] != "Dataset":
        raise ValueError(
            "Wrong object type for parquet archive. Expected Dataset, got"
            f" {data['__name__']}"
        )
    dataset = Dataset(
        images_root=data["images_root"],
        images=data["images"],
        annotations=data["annotations"],
        label_map=data["label_map"],
    )
    if "booleanized_columns" in data:
        dataset.booleanized_columns = data["booleanized_columns"]

    return dataset



from collections.abc import Callable, Iterable
from random import seed, shuffle
from typing import overload
from warnings import warn

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from ...utils.grouper import (
    ContinuousGroup,
    get_group_names,
    group_list,
    group_relational_data,
    groups_to_list,
)
from .balanced_groups import dataset_share_distance, df_to_hist, hist_distance
from .disjoint_groups import make_atomic_chunks


def get_winner(
    split_hists: pd.DataFrame | None,
    split_hists_distances: pd.Series | None,
    candidate_hist: pd.Series | None,
    split_sizes: pd.Series,
    candidate_size: int,
    hist_cost_function: Callable[[pd.Series], float],
    share_cost_function: Callable[[pd.Series], float],
    hist_cost_weight: float = 1,
    share_cost_weight: float = 1,
) -> tuple[str, pd.DataFrame | None, pd.Series | None, pd.Series]:
    """Get the best split i.e. with the lowest from series of precomputed costs.
    The series are histogram costs, i.e. with distribution distances for values the
    user which to be evenly distributed between splits, and the share costs the IOU
    distance between

    The result is then the key of the dictionary with the lowest consolidated cost.
    A special case is when all distribution costs are infinite. In that case, only
    consider the share cost.

    Args:
        split_hists: DataFrame containing the current histograms of splits. Columns are
            splits, and rows are histogram bins
        split_hists_distances: Series containing the cached distance values of distance
            between the split hist and the target histogram. If set to None, will
            recompute them
        candidate_hist: Series containing the histogram of the candidate atom. rows are
            the same as ``split_hists``
        split_sizes: Series containing the sizes of each split each row is a split.
        candidate_size: size of current atom. Depending on how the split is done, it's
            not necessary the same as the sum of candidate histogram.
        hist_cost_function: function that computes a score for a dataframe of
            histograms. This will be used to compute the histogram cost for each split
            if the atom was to be assigned to it.
        share_cost_function: function that computes a score for dataset repartition
            against a target split share. This is used to compute the cost of assigning
            the candidate atom to each split.
        hist_cost_weight: weight applied to histogram cost to choose the winner split.
            The higher, the more important the histogram cost will be for the decision.
            Defaults to 1.
        share_cost_weight: weight applied to share cost to choose the winner split.
            The higher, the more important the share cost will be for the decision.
            Defaults to 1.


    Returns:
        A tuple with 4 elements
         - name of the winning split
         - updated split histograms, as a DataFrame similar to ``split_hists``
           (None if given ``split_hists`` was None)
         - updated split hist costs, as a Series, similar to ``split_hists_distances``
           (None if given ``split_hists`` was None)
         - updated share of splits, as a Series, similar to ``split_shares``
    """
    # Construct aggregated split histogram cost Series, where each row is the split
    # we could assign the new atom, and the value is the corresponding cost that
    # We try to minimize
    split_names = split_sizes.index
    if split_hists is not None:
        assert candidate_hist is not None
        if split_hists_distances is None:
            split_hists_distances = split_hists.apply(hist_cost_function)
        updated_split_hists = split_hists.add(candidate_hist, axis="index")
        updated_split_hist_distances = updated_split_hists.apply(hist_cost_function)
        split_hist_distance_square = split_hists_distances.values
        ones = np.ones_like(split_hist_distance_square, dtype=int)
        split_hist_distance_square = split_hist_distance_square[:, None] @ ones[None]
        split_hist_distance_square[ones, ones] = updated_split_hist_distances
        aggregated_split_hists_costs = pd.Series(
            split_hist_distance_square.sum(axis=0), index=split_names
        )
    else:
        split_hists_distances = None
        updated_split_hist_distances = None
        updated_split_hists = None
        aggregated_split_hists_costs = pd.Series(0, index=split_names)

    # Construct aggregated share cost Series, whereas above, rows are split and
    # values are IoU between target share and the resulting share
    # if the atom was to be assigned to that split
    updated_sizes = split_sizes + candidate_size
    split_size_square = pd.DataFrame(
        split_sizes.values + np.diag(updated_sizes - split_sizes),
        index=split_names,
        columns=split_names,
    )

    aggregated_share_costs = split_size_square.apply(share_cost_function, axis=1)
    assert isinstance(aggregated_share_costs, pd.Series)

    infinite_histogram_cost = aggregated_split_hists_costs == float("inf")
    if all(infinite_histogram_cost):
        consolidated_cost = aggregated_share_costs
    else:
        consolidated_cost = (
            hist_cost_weight * aggregated_split_hists_costs
            + share_cost_weight * aggregated_share_costs
        )
    winner = consolidated_cost.idxmin()
    assert isinstance(winner, str)

    if split_hists is not None:
        split_hists[winner] = updated_split_hists[winner]  # pyright: ignore
        split_hists_distances[winner] = updated_split_hist_distances[  # pyright: ignore
            winner  # pyright: ignore
        ]

    split_sizes[winner] = updated_sizes[winner]
    return winner, split_hists, split_hists_distances, split_sizes


@overload
def split_dataframe(
    input_data: pd.DataFrame,
    root_data: pd.DataFrame,
    key_to_root: str = "image_id",
    input_seed: int = 0,
    split_names: Iterable[str] = ("train", "valid"),
    target_split_shares: Iterable[float] = (0.8, 0.2),
    split_column_name: str = "split",
    keep_separate_groups: group_list = ("image_id",),
    keep_balanced_groups: group_list = ("category_id",),
    keep_balanced_groups_weights: Iterable[float] | None = None,
    inplace: bool = False,
    split_at_root_level: bool = False,
    hist_cost_weight: float = 1,
    share_cost_weight: float = 1,
    earth_mover_regularization: float = 0,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    pass


@overload
def split_dataframe(
    input_data: pd.DataFrame,
    root_data: None = None,
    key_to_root: str = "image_id",
    input_seed: int = 0,
    split_names: Iterable[str] = ("train", "valid"),
    target_split_shares: Iterable[float] = (0.8, 0.2),
    split_column_name: str = "split",
    keep_separate_groups: group_list = ("image_id",),
    keep_balanced_groups: group_list = ("category_id",),
    keep_balanced_groups_weights: Iterable[float] | None = None,
    inplace: bool = False,
    split_at_root_level: bool = False,
    hist_cost_weight: float = 1,
    share_cost_weight: float = 1,
    earth_mover_regularization: float = 0,
) -> pd.DataFrame:
    pass


@overload
def split_dataframe(
    input_data: pd.DataFrame,
    root_data: pd.DataFrame | None = None,
    key_to_root: str = "image_id",
    input_seed: int = 0,
    split_names: Iterable[str] = ("train", "valid"),
    target_split_shares: Iterable[float] = (0.8, 0.2),
    split_column_name: str = "split",
    keep_separate_groups: group_list = ("image_id",),
    keep_balanced_groups: group_list = ("category_id",),
    keep_balanced_groups_weights: Iterable[float] | None = None,
    inplace: bool = False,
    split_at_root_level: bool = False,
    hist_cost_weight: float = 1,
    share_cost_weight: float = 1,
    earth_mover_regularization: float = 0,
) -> pd.DataFrame | tuple[pd.DataFrame, pd.DataFrame]:
    pass


def split_dataframe(
    input_data: pd.DataFrame,
    root_data: pd.DataFrame | None = None,
    key_to_root: str = "image_id",
    input_seed: int = 0,
    split_names: Iterable[str] = ("train", "valid"),
    target_split_shares: Iterable[float] = (0.8, 0.2),
    split_column_name: str = "split",
    keep_separate_groups: group_list = ("image_id",),
    keep_balanced_groups: group_list = ("category_id",),
    keep_balanced_groups_weights: Iterable[float] | None = None,
    inplace: bool = False,
    split_at_root_level: bool = False,
    hist_cost_weight: float = 1,
    share_cost_weight: float = 1,
    earth_mover_regularization: float = 0,
) -> pd.DataFrame | tuple[pd.DataFrame, pd.DataFrame]:
    """Perform the split operation on input_data and root_data.

    This algorithm works in 2 steps:

    1. divide the dataframe into atomic sub frames. Given the image and annotation
        attributes that need to be kept separate, we can construct sub frame of elements
        that cannot be in different splits.
    2. Construct the split dataframes iteratively by trying to keep given column values
        with a balanced repartition between splits, along with keeping split sizes as
        close to target share as possible. Each atomic sub frame is routed to the split
        that minimize a cost function which try to optimize repartition targets.

    Args:
        input_data: DataFrame containing input_data information, must contain at least
            the column given in ``key_to_root``.
        root_data: DataFrame containing image information. its index must contain all
            values contained in the ``image_id`` column of the input_data DataFrame.
        key_to_root: name of the column in input that refers to id in root data
            dataframe. Defaults to "image_id".
        input_seed: Seed used for shuffling sub frames before beginning step 2 of
            splitting algorithm. Defaults to 0.
        split_names: Names of splits. Must be the same length as
            ``target_split_shares``. Defaults to ("train", "valid").
        target_split_shares: List of relative size of each split. Must be the same
            length as ``split_names``, and will be normalized so that its sum is 1.
            Defaults to (0.8, 0.2).
        split_column_name: Name of the column where the split value of dataset will be
            read and written. Defaults to "split".
        keep_separate_groups: columns in ``input_data`` or `root_data``
            DataFrame to keep separate. That is for a particular column, two rows with
            the same value cannot be in different splits. Defaults to ("image_id",).
        keep_balanced_groups: columns or groups (as defined in
            ``input_data`` or ``root_data`` DataFrames to keep balanced.
            That is for a particular column, the distribution of values is the same
            between original DataFrame and its split, as much as possible.
            Defaults to ("category_id",).
        keep_balanced_groups_weights: Importance of each group to keep balanced when
            computing histogram cost. If not None, must be of the same size as
            ``keep_separate_groups``. Defaults to None.
        inplace: If set, will modify dataframes inplace. This can silently modify some
            objects (like Datasets) that use them. Defaults to False.
        split_at_root_level: If set, will compute split sizes (and thus share distances)
            at root level, i.e. regarding sizes in the ``root_data`` dataframe. As a
            consequence, the split column name will be added to
            ``keep_separate_input_groups`` if it's not already in it, and the number of
            rows in the input data per row in root data will not have any influence on
            the share cost.
        hist_cost_weight: importance of histogram cost for balanced groups.
            The higher, the more important the histogram cost will be for the decisio of
            where to put each split. Defaults to 1.
        share_cost_weight: importance of share cost for balanced groups.
            The higher, the more important the share cost will be for the decision of
            where to put each split. Defaults to 1.
        earth_mover_regularization: Regularization parameter applied to sinkhorn's
            algorithm during earth mover distance computation. See
            :func:`.earth_mover_distance`. Defaults to 0

    Returns:
        new annotation and root_data with the split column populated with the
        corresponding split name.
    """
    target_split_shares = pd.Series(list(target_split_shares), index=[*split_names])
    target_split_shares /= target_split_shares.sum()

    keep_balanced_groups = groups_to_list(keep_balanced_groups)
    for g in keep_balanced_groups:
        if isinstance(g, ContinuousGroup) and g.label_type == "intervals":
            g.label_type = "mid"
    keep_balanced_group_names = get_group_names(keep_balanced_groups)
    keep_separate_groups = groups_to_list(keep_separate_groups)

    if keep_balanced_groups_weights is None:
        keep_balanced_groups_weights = pd.Series(
            1, index=keep_balanced_group_names, dtype=float
        )
    else:
        keep_balanced_groups_weights = pd.Series(
            list(keep_balanced_groups_weights),
            index=keep_balanced_group_names,
            dtype=float,
        )

    if not inplace:
        input_data = input_data.copy()
        root_data = root_data.copy() if root_data is not None else None

    if split_at_root_level:
        assert key_to_root in input_data.columns
    if root_data is not None:
        assert key_to_root in input_data.columns
        assert input_data[key_to_root].isin(root_data.index).all()
        if split_column_name in root_data.columns:
            input_data[split_column_name] = root_data.loc[
                input_data[key_to_root], split_column_name
            ].values
        if split_at_root_level and key_to_root not in keep_separate_groups:
            keep_separate_groups.append(key_to_root)

    if split_column_name in input_data:
        already_assigned = input_data[split_column_name].isin(split_names)
        if already_assigned.sum() == len(input_data):
            warn(
                "Every row in the DataFrame is already assigned to a given split. It's"
                " possible you forgot to remove the already existing split values"
                " before splitting ",
                RuntimeWarning,
            )

    keep_separate_input_groups_dict, *_ = group_relational_data(
        input_data,
        keep_separate_groups,
        root_data,
        key_to_root=key_to_root,
    )
    keep_separate_input_pandas_groups = list(keep_separate_input_groups_dict.values())

    print("Separating input data into atomic chunks")
    atomic_chunks, assigned_chunks = make_atomic_chunks(
        data=input_data,
        groups=keep_separate_input_pandas_groups,
        split_column=split_column_name,
        split_names=target_split_shares.index,  # pyright: ignore
    )
    if not atomic_chunks:
        print("No chunk to distribute")
    else:
        print(
            f"{len(atomic_chunks)} chunks to distribute"
            f" across {len(target_split_shares)} splits"
        )

    # Construct a split dictionary, containing the indexes of input_data, belonging
    # to each split
    splits: dict[str, list] = {str(name): [] for name in target_split_shares.index}
    split_sizes = pd.Series(0, index=target_split_shares.index)

    def share_cost_function(candidate_shares: pd.Series) -> float:
        return dataset_share_distance(target_split_shares, candidate_shares)

    (
        keep_balanced_groups_dict,
        category_groups,
        continuous_groups,
    ) = group_relational_data(
        input_data,
        keep_balanced_groups,
        root_data,
        key_to_root=key_to_root,
    )

    category_weights = keep_balanced_groups_weights[
        keep_balanced_groups_weights.index.isin(category_groups)
    ]
    continuous_weights = keep_balanced_groups_weights[
        keep_balanced_groups_weights.index.isin(continuous_groups)
    ]

    keep_balanced_pandas_groups = list(keep_balanced_groups_dict.values())
    if keep_balanced_pandas_groups:
        target_hist = df_to_hist(input_data, keep_balanced_pandas_groups)
        # Construct the histogram of each split that will be updated along with
        # their construction. split_hists is a dataframe of histograms.
        # Each column is a split
        split_hists = pd.DataFrame(
            0, index=target_hist.index, columns=target_split_shares.index
        )

        # Function that we will apply to the split histogram dataframe
        def hist_cost_function(split_hist: pd.Series) -> float:
            return hist_distance(
                target_hist,
                split_hist,
                category_weights,
                continuous_weights,
                sinkhorn_lambda=earth_mover_regularization,
            )

    else:
        target_hist = None
        split_hists = None

        def hist_cost_function(split_hist: pd.Series) -> float:
            return 0.0

    # Already assigned chunks belong to a split, so we add them to the split's index
    # list and update their histogram
    for name, group in assigned_chunks.items():
        if name in split_names:
            splits[name].append(group.index)
            if target_hist is not None:
                assert split_hists is not None
                split_hists[name] += df_to_hist(
                    group,
                    keep_balanced_pandas_groups,
                    full_index=target_hist.index,
                )
            split_sizes[name] += len(group)
        else:
            atomic_chunks.append(group)

    seed(input_seed)
    shuffle(atomic_chunks)

    if target_hist is not None:
        assert split_hists is not None
        candidate_hists = [
            df_to_hist(
                atom,
                keep_balanced_pandas_groups,
                full_index=target_hist.index,
            )
            for atom in atomic_chunks
        ]
        # Construct the distances between each split histogram and the target histogram
        # If the splits are not populated, we should get infinity everywhere
        split_hists_distances = split_hists.apply(hist_cost_function)
    else:
        candidate_hists = None
        split_hists_distances = None
    # For each atom, decide where it will go by computing the overall distance between
    # each split histogram and the target histogram, under the hypothesis of going
    # to a particular split. The winner split is the hypothesis with the lowest score,
    # gets the atom and its histogram gets updated
    for i, atom in enumerate(tqdm(atomic_chunks, disable=len(atomic_chunks) == 0)):
        if len(atom) == 0:
            continue
        if candidate_hists is not None:
            current_candidate_hist = candidate_hists[i]
        else:
            current_candidate_hist = None
        if split_at_root_level:
            candidate_size = len(atom[key_to_root].unique())
        else:
            candidate_size = len(atom)
        winner, split_hists, split_hists_distances, split_sizes = get_winner(
            split_hists,
            split_hists_distances,
            current_candidate_hist,
            split_sizes,
            candidate_size,
            hist_cost_function,
            share_cost_function,
            hist_cost_weight,
            share_cost_weight,
        )
        splits[winner].append(atom.index)

    for name, split in splits.items():
        if not split:
            continue
        input_data_to_mark = np.concatenate(split)
        input_data.loc[input_data_to_mark, split_column_name] = name
        if root_data is not None and split_at_root_level:
            root_data_to_mark = input_data.loc[
                input_data[split_column_name] == name, key_to_root
            ].unique()
            root_data.loc[root_data_to_mark, split_column_name] = name

    if root_data is not None:
        return input_data, root_data
    else:
        return input_data



from collections.abc import Iterable
from itertools import combinations
from pathlib import Path

import pandas as pd
import xmltodict
from tqdm import tqdm

from lours.dataset import Dataset
from lours.utils.bbox_converter import import_bbox


def from_pascalVOC_generic(
    annotations_root: Path,
    images_root: Path | str | None = None,
    split_folder: Path | str | None = None,
    split_values: Iterable[str] | str = ("train", "val"),
) -> Dataset:
    """Load a dataset in pascalVOC format

    See `specifications`__ (only Object detection)

        .. __: http://host.robots.ox.ac.uk/pascal/VOC/voc2012/htmldoc/devkit_doc.html

    See Also:
        :func:`.from_pascalVOC_detection`

    Notes:
        - This has been tested against PascalVOC2012
        - If loading official detection splits, not all images will be assigned a split
          value.
        - For objects with "parts" (like hands for persons), a new object will be
          created and a ``body_id`` column will link to the corresponding root object.
        - ``actions.<value>`` columns are converted to boolean and included in the
          booleanized column ``actions``.
        - ``difficult``, ``truncated`` and ``occluded`` columns are converted to
          boolean.

    Args:
        annotations_root: Folder containing the xml files containing annotations
        images_root: Folder containing the image files. Path to images are given in the
            corresponding annotation files, relative to this folder. If set to None,
            will be assumed to be the same as ``annotations_root``. Defaults to None.
        split_folder: Folder containing txt file for each split. Files are named
            ``<split>.txt`` and contain all the file name without extension contained
            in that split. If set to None, will be assumed to be the same as
            ``annotations_root``. Defaults to None.
        split_values: split value or list of split values to read. Following the
            aforementioned syntax, will try to open the corresponding split files and
            assign this split value to the corresponding images. Note that split values
            need to be exclusive to each other. For example, you cannot load both
            "train" and "trainval" splits. Defaults to ("train", "val").

    Returns:
        Loaded dataset with split values assigned
    """
    if images_root is None:
        images_root = annotations_root
    if split_folder is None:
        split_folder = annotations_root
    annotations_root = Path(annotations_root)
    images_root = Path(images_root)
    split_folder = Path(split_folder)
    if isinstance(split_values, str):
        split_values = [split_values]
    xml_files = list(annotations_root.glob("**/*.xml"))

    def image_set(split: str) -> set[str]:
        try:
            with open(split_folder / f"{split}.txt") as f:
                split_images = f.read().strip().split("\n")
            return set(split_images)
        except FileNotFoundError:
            return set()

    split_files: dict[str, set[str]] = {}
    for value in split_values:
        split_files[value] = image_set(value)

    for s1, s2 in combinations(split_files.keys(), 2):
        overlap = split_files[s1].intersection(split_files[s2])
        assert not overlap, f"Splits {s1} and {s2} have non null overlap : {overlap}"

    reversed_splits = {}
    for split_name, split_set in split_files.items():
        for file_name in split_set:
            reversed_splits[file_name] = split_name
    annotations_dicts = []

    image_data_list = {}

    current_object_id = 0

    for image_id, xml_file in enumerate(tqdm(xml_files)):
        annotation_dict = xmltodict.parse(xml_file.read_text())["annotation"]

        image_data = {}
        image_data["relative_path"] = annotation_dict["filename"]
        image_data["width"] = int(annotation_dict["size"]["width"])
        image_data["height"] = int(annotation_dict["size"]["height"])
        image_data["split"] = reversed_splits.get(xml_file.stem, None)

        image_data_list[image_id] = image_data

        objects = annotation_dict["object"]
        if isinstance(objects, dict):
            objects = [objects]
        object_parts = []
        for object in objects:
            object["id"] = current_object_id
            object["image_id"] = image_id
            if "part" in object:
                parts = object.pop("part")
                if isinstance(parts, dict):
                    parts = [parts]
                for p in parts:
                    p["body_id"] = current_object_id
                object_parts.extend(parts)
            if "point" in object:
                point = object.pop("point")
                x, y = point["x"], point["y"]
                object_parts.append(
                    {
                        "bndbox": {"xmax": x, "xmin": x, "ymax": y, "ymin": y},
                        "image_id": image_id,
                        "body_id": current_object_id,
                        "name": "person of interest",
                    }
                )
            current_object_id += 1
        for part in object_parts:
            part["id"] = current_object_id
            part["image_id"] = image_id
            current_object_id += 1
        annotations_dicts.extend(objects)
        annotations_dicts.extend(object_parts)
    images_df = pd.DataFrame.from_dict(image_data_list, orient="index")
    images_df.index = images_df.index.rename("id")
    annotations_df = pd.json_normalize(annotations_dicts).set_index("id")

    annotations_df = annotations_df.astype(
        {"body_id": pd.Int64Dtype()}, errors="ignore"
    )
    action_columns = [
        name for name in annotations_df.columns if name.startswith("actions.")
    ]
    annotations_df[action_columns] = (
        annotations_df[action_columns]
        .replace({"0": False, "1": True})
        .fillna(False)
        .astype(bool)
    )

    to_boolean = list(
        {"difficult", "occluded", "truncated"}.intersection(set(annotations_df.columns))
    )
    annotations_df[to_boolean] = (
        annotations_df[to_boolean]
        .replace(
            {
                "0": False,
                "1": True,
            }
        )
        .astype(pd.BooleanDtype)
    )

    bounding_boxes = import_bbox(
        annotations_df[["bndbox.xmin", "bndbox.xmax", "bndbox.ymin", "bndbox.ymax"]]
        .astype(float)
        .to_numpy(),
        images_df,
        image_ids=annotations_df["image_id"],
        input_format="XXYY",
    )

    annotations_df = pd.concat(
        [
            annotations_df.drop(
                [
                    "bndbox.xmin",
                    "bndbox.xmax",
                    "bndbox.ymin",
                    "bndbox.ymax",
                ],
                axis=1,
            ),
            bounding_boxes,
        ],
        axis=1,
    )

    label_map = dict(enumerate(annotations_df["name"].unique()))
    reverse_label_map = {v: k for k, v in label_map.items()}
    annotations_df["category_id"] = annotations_df["name"].replace(reverse_label_map)
    annotations_df = annotations_df.rename(columns={"name": "category_str"})

    dataset = Dataset(
        images_root=images_root,
        images=images_df,
        annotations=annotations_df,
        label_map=label_map,
    )

    dataset.booleanized_columns["annotations"] = {"actions"}

    return dataset


def from_pascalVOC_detection(input_folder: Path) -> Dataset:
    """Load a pascalVOC detection dataset that follows the official structure.

    Folder is assumed to contain three sub-folders:

    - "Annotations" containing the annotation xml files
    - "JPEGImages" containing the images files
    - "ImageSets/Main" containing the detection split files

    See `specifications`__

        .. __: http://host.robots.ox.ac.uk/pascal/VOC/voc2012/htmldoc/devkit_doc.html

    See Also:
        :func:`.from_pascalVOC_generic`

    Notes:
        - This has been tested against PascalVOC2012
        - If loading official detection splits, not all images will be assigned a split
          value.
        - For objects with "parts" (like hands for persons), a new object will be
          created and a ``body_id`` column will link to the corresponding root object.
        - ``actions.<value>`` columns are converted to boolean and included in the
          booleanized column ``actions``.
        - ``difficult``, ``truncated`` and ``occluded`` columns are converted to
          boolean.
        - The dataset will remove images without split. If you wish to load all images
          with available annotation, use :func:`.from_pascalVOC_generic`.

    Args:
        input_folder: Folder containing annotations, images, and split folders.

    Returns:
        Loaded dataset
    """
    annotations_folder = input_folder / "Annotations"
    images_root = input_folder / "JPEGImages"
    split_folder = input_folder / "ImageSets" / "Main"
    pascal_dataset = from_pascalVOC_generic(
        annotations_folder, images_root, split_folder
    )
    # Remove images that neither in the train nor in the valid split
    return pascal_dataset.loc[~pascal_dataset.images["split"].isna()]



import warnings
from collections import defaultdict
from collections.abc import Iterable, Sequence
from dataclasses import dataclass

import pandas as pd


@dataclass
class IndexedSet:
    """Class representing a set with a corresponding list of indexes indicating what
    sets in the initial list where used to construct this one. In other word, there's
    an original list of sets and the union of all sets indexed make up the current
    set.
    """

    index: set[int]
    """index $i$ of sets $S_i$ that were used when constructing this set.
    """

    merged_set: set
    r"""
    Resulting set.

    .. math::
        S = \bigcup_{i \in \text{index}} S_i
    """

    def union(self, *others: "IndexedSet") -> "IndexedSet":
        r"""Perform the union operation. union operation is applied on both index sets
        and the sets themselves.

        Args:
            *others: Iterable of $n$ other indexed sets :math:`(S_i, \text{index}_i)`
                to perform the union operation

        Returns:
            new indexed set with

            .. math::

                \text{index} &= \text{index}_1 \cup \text{index}_2 \cup \cdots
                  \cup \text{index_n} \\
                S &= S_1 \cup S_2 \cup \cdots \cup S_n
        """
        return IndexedSet(
            self.index.union(*[o.index for o in others]),
            self.merged_set.union(*[o.merged_set for o in others]),
        )

    def is_disjoint(self, other: "IndexedSet") -> bool:
        """Tell if the intersection between current index set and another one is empty
        or not

        Args:
            other: other indexed set that we want the intersection with

        Returns:
            True if intersection is empty, False otherwise
        """
        return self.merged_set.intersection(other.merged_set) == set()


def factorize_sets(input_sets: Sequence[set]) -> list[list[int]]:
    r"""From an index-able sequence of sets, partition all possible values in factor
    sets so that two elements in a particular factor set can be linked with a sequence
    of input sets with a non-null intersection.


    .. math::

        \widehat{S} = \bigcup_i S_i \in input sets

        \forall x,y \in \widehat{S} , \exists i_0 , i_1, \cdots , i_n,
          x \in S_{i_0}, y \in S_{i_n}, \forall j,
          S_{i_j} \cap S_{i_{j+1}} \neq \emptyset


    Args:
        input_sets: sequence of sets with possible overlapping values that need to be
            factorized.


    Returns:
       list of set indices for each factor. That is, the index in the input sets
       sequence to recreate the factor sets with a union operation.
    """
    indexed_input_sets = [
        IndexedSet({i}, current_set) for i, current_set in enumerate(input_sets)
    ]
    merged_indices = []
    while indexed_input_sets:
        first_id_set, *remaining = indexed_input_sets
        to_merge, to_keep = [], []
        for id_set in remaining:
            (
                to_keep.append(id_set)
                if first_id_set.is_disjoint(id_set)
                else to_merge.append(id_set)
            )
        if not to_merge:
            merged_indices.append(first_id_set.index)
            indexed_input_sets = remaining
            continue
        indexed_input_sets = [first_id_set.union(*to_merge)] + to_keep
    return merged_indices


def give_already_assigned(
    data: pd.DataFrame, split_column: str = "split", split_names: Iterable[str] = ()
) -> tuple[list[pd.DataFrame], dict[str, pd.DataFrame]]:
    """Divide a DataFrame with a split column into chunks with an assigned split and
    unassigned chunks. Unassigned chunks are chunks with an invalid split values
    (like Nan or None) or split values that are not in the list ``split_names``

    Args:
        data: input DataFrame to divide
        split_column: name of the split column. Defaults to "split".
        split_names: list of allowed split names. If the split value is not in it, the
            group is considered unassigned

    Returns:
        tuple with 2 elements
         - list of unassigned DataFrame groups
         - dictionary of assigned DataFrame groups where key is the split name
    """
    split_names = [*split_names]
    if split_column not in data.columns:
        return [df for _, df in data.groupby(level=0, as_index=False, dropna=False)], {}

    unassigned = []
    assigned = {}
    for split_name, df in data.groupby(split_column, dropna=False):
        if split_name in split_names:
            assigned[split_name] = df
        else:
            unassigned.extend([row for _, row in df.groupby(level=0, dropna=False)])
    return unassigned, assigned


def make_atomic_chunks(
    data: pd.DataFrame,
    groups: Iterable[str | pd.Series],
    split_column: str = "split",
    split_names: Iterable[str] = (),
) -> tuple[list[pd.DataFrame], dict[str, pd.DataFrame]]:
    r"""Subdivide the input DataFrame into dissociate chunks from given columns.
    In other words, for two rows in distinct chunks, there will never be the same
    elements in the involved columns, and for two rows in the same chunk, there can be a
    chain of elements all in this chunk to link them. For example, $(A, B)$ and $(C, D)$
    have different values for each column, but if there exist a row $(A, D)$, then we
    can make the chain :math:`(A, B) \rightarrow (A, D) \rightarrow (C, D)`, which
    means the three rows will be in the same chunk.

    Note:
        In the case the data has a ``split`` column with non NaN values, the
        corresponding rows and the chunk they are linked to will be completely assigned
        to that split. However, it will raise an error if a theoretically indivisible
        chunk has rows with different split values.

    Args:
        data: DataFrame to be split into dissociated chunks.
        groups: groups to consider for the dissociation. If group is a string, given
            DataFrame in ``data`` must include a column with this name. If groups is a
            pandas categorical Series, given DataFrame in ``data`` must have the same
            index.
        split_column: Name of the column in ``data`` where the split value will be
            grabbed from. Rows with values within ``split_names`` will be considered
            assigned.
        split_names: Names of wanted splits. rows with split values outside of it will
            be considered unassigned.

    Returns:
        1. List of DataFrames corresponding to the dissociated chunks. concatenating the
           returned DataFrames would end up in the input DataFrame.
        2. dictionary with already assigned atomic chunk, because the "split" value was
           already filled in at least one of the rows
    """
    if not groups:
        return give_already_assigned(data, split_column, split_names)
    else:
        # Reorder groups so that the first group to use groupby method on is the one
        # with the smallest number of unique values
        def group_sort_key(group):
            if isinstance(group, str):
                return len(data[group].unique())
            else:
                return len(group.unique())

        groups = sorted(groups, key=group_sort_key)
        df_list = [
            df
            for _, df in data.groupby(
                groups[0], as_index=False, dropna=False, observed=True
            )
        ]
        if len(groups) > 1:
            for g in groups[1:]:
                unique_values = [
                    set(df[g].unique()) if isinstance(g, str) else set(g.unique())
                    for df in df_list
                ]
                clusters = factorize_sets(unique_values)
                new_df_list = []
                for cluster in clusters:
                    new_df_list.append(pd.concat([df_list[i] for i in cluster]))
                df_list = new_df_list
    unassigned = []
    already_assigned = defaultdict(list)
    if split_column in data.columns:
        for df in df_list:
            split_names = [
                name
                for name in df[split_column].dropna().unique()
                if name in split_names
            ]
            if len(split_names) > 1:
                split_names_str = ", ".join(map(str, split_names))
                warnings.warn(
                    "One chunk has multiple split assignments"
                    f" ({split_names_str}) and will be treated as unassigned",
                    RuntimeWarning,
                )
                unassigned.append(df)
            elif len(split_names) == 1:
                split_name = split_names[0]
                df["split"] = split_name
                already_assigned[split_name].append(df)
            else:
                unassigned.append(df)
    else:
        unassigned = df_list
    already_assigned = {
        name: pd.concat(split) for name, split in already_assigned.items()
    }
    return unassigned, already_assigned
